# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""debusine-admin to perform storage maintenance."""

import datetime as dt
from collections.abc import Callable
from pathlib import Path
from typing import Any, NoReturn, override

import pgtransaction
from django.conf import settings
from django.core.management import CommandParser
from django.db import transaction
from django.db.models import F, Max, OuterRef, Q, Subquery
from django.utils import timezone

from debusine.db.models import (
    Artifact,
    ArtifactRelation,
    CollectionItem,
    File,
    FileInStore,
    FileStore,
    FileStoreInScope,
    FileUpload,
    Scope,
)
from debusine.db.transactions import atomic
from debusine.django.management.debusine_base_command import DebusineBaseCommand
from debusine.server.file_backend.interface import FileBackendEntryInterface


class Command(DebusineBaseCommand):
    """Command to report and fix inconsistencies."""

    help = "Report and fix inconsistencies in Debusine's DB and storage"

    @override
    def add_arguments(self, parser: CommandParser) -> None:
        parser.add_argument(
            "--dry-run",
            help="Trial run: do not modify anything",
            action="store_true",
        )

    @override
    def handle(self, *args: Any, **options: Any) -> NoReturn:
        """Run the checks and fixes."""
        dry_run = options["dry_run"]
        with transaction.atomic():
            self._delete_unreferenced_files(dry_run=dry_run)
        self._delete_empty_directories_in_the_store_directory(dry_run=dry_run)
        storage_policies_correct = self._apply_storage_policies(dry_run=dry_run)
        with transaction.atomic():
            self._delete_incomplete_old_artifacts(dry_run=dry_run)
        with transaction.atomic():
            self._delete_redundant_uploads(dry_run=dry_run)
        with transaction.atomic():
            self._delete_orphan_files_from_upload_directory(dry_run=dry_run)
        with transaction.atomic():
            missing_files_from_stores_correct = (
                self._report_missing_files_stores_referenced_by_the_db()
            )
        with transaction.atomic():
            missing_files_from_upload_correct = (
                self._report_missing_files_referenced_by_db_file_upload()
            )
        with atomic(isolation_level=pgtransaction.REPEATABLE_READ, retry=0):
            self._roll_up_file_store_journal(dry_run=dry_run)

        raise SystemExit(
            int(
                not storage_policies_correct
                or not missing_files_from_stores_correct
                or not missing_files_from_upload_correct
            )
        )

    def _delete_unreferenced_files(self, *, dry_run: bool) -> None:
        """Delete files in stores that are not referenced by the DB."""
        self.print_verbose("Checking orphan files from stores")
        one_day_ago = timezone.now() - dt.timedelta(days=1)

        # We consider only files older than a day old in order to allow a
        # grace period for files that are in the process of being created,
        # to avoid possible race conditions where this job deletes them
        # while we're still in the middle of creating references to them.
        # TODO: At the moment we can only get mtimes from some backends,
        # although this isn't an inherent limitation.
        for store in FileStore.objects.filter(
            backend__in={
                FileStore.BackendChoices.LOCAL,
                FileStore.BackendChoices.S3,
            }
        ):
            established_entries_in_storage = set(
                store.get_backend_object().list_entries()
            )
            entries_in_db = self._entries_from_files_in_file_store_db(store)
            entries_to_delete = established_entries_in_storage - entries_in_db

            for entry_to_delete in entries_to_delete:
                mtime = entry_to_delete.get_mtime()
                if mtime is not None and mtime < one_day_ago:
                    if not dry_run:
                        entry_to_delete.remove()

                    self.print_verbose(f"Deleted {entry_to_delete}")

    @staticmethod
    def _path_older_than_one_day(path: Path) -> bool:
        one_day_ago = dt.datetime.now(tz=dt.UTC) - dt.timedelta(days=1)
        file_modification_time = dt.datetime.fromtimestamp(
            path.stat().st_mtime, tz=dt.UTC
        )
        return file_modification_time < one_day_ago

    def _delete_empty_directories_in_the_store_directory(
        self, *, dry_run: bool
    ) -> None:
        """
        Delete empty directories in the store directory.

        This method might leave new empty directories (by design).

        If the directory a/b/c is deleted: now the directory a/b might be empty.
        This method could then check if a/b is empty but:

        - mtime of a/b is the recent date
        - if it deleted a/b it could be that the LocalFileBackend is adding
           a file in a/b and could fail (since LocalFileBackend is creating
           the subdirectories incrementally).

        To fix this it could have a locking mechanism (DB based or file
        based).

        For now: empty directories might be eventually deleted but not
        "right now".
        """
        self.print_verbose("Checking empty directories")
        store = Path(settings.DEBUSINE_STORE_DIRECTORY)

        for dirpath, dirnames, filenames in store.walk():
            if (
                not dirnames
                and not filenames
                and self._path_older_than_one_day(dirpath)
                and dirpath != store
            ):
                # The directory is empty (and was modified more than a day
                # ago): it can be deleted
                if not dry_run:
                    try:
                        dirpath.rmdir()
                    except OSError:
                        # Probably a file has been added between
                        # _list_directory() call and the rmdir()
                        continue
                self.print_verbose(f"Deleted empty directory {dirpath}")

    def _copy_file(
        self,
        *,
        scope: Scope,
        source: FileStore,
        target: FileStore,
        file: File,
        move: bool,
        dry_run: bool,
    ) -> None:
        """
        Copy a file between two file stores.

        If `move` is True, then also remove the file from the source.
        """
        if dry_run:
            verb = "Would move" if move else "Would copy"
        else:
            source_backend = source.get_backend_object()
            target_backend = target.get_backend_object()
            with source_backend.get_temporary_local_path(file) as local_path:
                target_backend.add_file(local_path, file)
            if move:
                source_backend.remove_file(file)
            verb = "Moved" if move else "Copied"

        self.print_verbose(
            f"{scope}: {verb} {file.hash_digest.hex()} from {source.name} "
            f"to {target.name}"
        )

    def _remove_file(
        self, *, scope: Scope, store: FileStore, file: File, dry_run: bool
    ) -> None:
        """Remove a file from a file store."""
        if dry_run:
            verb = "Would remove"
        else:
            store.get_backend_object().remove_file(file)
            verb = "Removed"

        self.print_verbose(
            f"{scope}: {verb} {file.hash_digest.hex()} from {store.name}"
        )

    def _drain_file(
        self,
        *,
        scope: Scope,
        policies: FileStoreInScope,
        store: FileStore,
        file: File,
        dry_run: bool,
    ) -> bool:
        """Drain a file from a file store in whatever way makes sense."""
        if scope.download_file_stores(file).exclude(id=store.id).exists():
            # The file can already be downloaded from some other store, so
            # we just need to remove it from this one.
            self._remove_file(
                scope=scope, store=store, file=file, dry_run=dry_run
            )
        else:
            # We can't easily optimize this by doing it in bulk for all
            # files, as the answer may change if stores exceed their
            # soft_max_size or max_size.
            upload_stores = scope.upload_file_stores(
                file, enforce_soft_limits=True
            ).exclude(id=store.id)
            if policies.drain_to is not None:
                upload_stores = upload_stores.filter(name=policies.drain_to)
            upload_store = upload_stores.first()
            if upload_store is None:
                if policies.drain_to is not None:
                    self.stderr.write(
                        f"{scope}: Wanted to drain {store.name} of "
                        f"{file.hash_digest.hex()}, but {policies.drain_to} "
                        f"cannot accept it"
                    )
                else:
                    self.stderr.write(
                        f"{scope}: Wanted to drain {store.name} of "
                        f"{file.hash_digest.hex()}, but it has no other upload "
                        f"file stores"
                    )
                return False

            self._copy_file(
                scope=scope,
                source=store,
                target=upload_store,
                file=file,
                move=True,
                dry_run=dry_run,
            )

        return True

    def _apply_populate_policy(self, scope: Scope, *, dry_run: bool) -> bool:
        """
        Apply per-scope `populate` file storage policy.

        This may involve copying files between stores.
        """
        ok = True

        for store in scope.file_stores.filter(
            filestoreinscope__populate=True
        ).order_by("id"):
            policies = scope.filestoreinscope_set.get(file_store=store)
            # Enforced by a check constraint on FileStoreInScope.
            assert not policies.drain
            assert not policies.read_only

            files_to_populate = (
                File.objects.filter(artifact__workspace__scope=scope)
                .exclude(filestore=store)
                .order_by("sha256")
            )
            if files_to_populate.exists():
                self.print_verbose(f"{scope}: Populating {store.name}")

            for file in files_to_populate.distinct().annotate(
                # Similar to Scope.download_file_stores, but in bulk and
                # excluding the store being populated.
                download_store_id=Subquery(
                    FileStore.objects.exclude(id=store.id)
                    .exclude(filestoreinscope__write_only=True)
                    .filter(files=OuterRef("id"))
                    .order_by(
                        F("filestoreinscope__download_priority").desc(
                            nulls_last=True
                        ),
                        F("filestoreinscope__upload_priority").desc(
                            nulls_last=True
                        ),
                    )
                    .values("id")[:1]
                )
            ):
                # Handle each file in a separate transaction; population may
                # take some time, and we don't want to have to wait for the
                # entire store to be populated before files added to it are
                # visible, or to roll back all the new FileInStore rows
                # because one copy failed.
                with transaction.atomic():
                    # Because this is in a separate transaction from the
                    # query that found files to populate, it's possible that
                    # the file has since been deleted.  If so, we can ignore
                    # it.
                    try:
                        file.refresh_from_db()
                    except File.DoesNotExist:
                        continue

                    if file.download_store_id is None:
                        self.stderr.write(
                            f"{scope}: Wanted to populate {store.name} with "
                            f"{file.hash_digest.hex()}, but it has no other "
                            f"download file stores"
                        )
                        ok = False
                        continue

                    if (
                        not scope.upload_file_stores(
                            file,
                            enforce_soft_limits=True,
                            include_write_only=True,
                        )
                        .filter(id=store.id)
                        .exists()
                    ):
                        # This might be slightly premature (we could have
                        # run into an unusually large file, and could
                        # perhaps populate some smaller files), but stores
                        # with the `populate` policy set ultimately need to
                        # have enough capacity to hold everything anyway.
                        self.stderr.write(
                            f"{scope}: Stopping population of {store.name} to "
                            f"avoid exceeding size limits"
                        )
                        ok = False
                        break

                    download_store = FileStore.objects.get(
                        id=file.download_store_id
                    )
                    self._copy_file(
                        scope=scope,
                        source=download_store,
                        target=store,
                        file=file,
                        move=False,
                        dry_run=dry_run,
                    )

        return ok

    def _apply_drain_policy(self, scope: Scope, *, dry_run: bool) -> bool:
        """
        Apply per-scope `drain` file storage policy.

        This may involve moving files between stores.
        """
        ok = True

        for store in scope.file_stores.filter(
            filestoreinscope__drain=True
        ).order_by("id"):
            policies = scope.filestoreinscope_set.get(file_store=store)
            # Enforced by a check constraint on FileStoreInScope.
            assert not policies.populate

            files_to_drain = File.objects.filter(
                artifact__workspace__scope=scope, filestore=store
            ).order_by("sha256")
            if files_to_drain.exists():
                self.print_verbose(f"{scope}: Draining {store.name}")

            for file in files_to_drain.distinct():
                # Handle each file in a separate transaction; draining may
                # take some time, and we don't want to have to wait for the
                # entire store to be drained before files moved from it to
                # other stores are visible, or to roll back all the new
                # FileInStore rows because one move failed.
                with transaction.atomic():
                    # Because this is in a separate transaction from the
                    # query that found files to drain, it's possible that
                    # the file has since been deleted.  If so, we can ignore
                    # it.
                    try:
                        file.refresh_from_db()
                    except File.DoesNotExist:
                        continue

                    if not self._drain_file(
                        scope=scope,
                        policies=policies,
                        store=store,
                        file=file,
                        dry_run=dry_run,
                    ):
                        ok = False

        return ok

    def _apply_max_size_limits(self, scope: Scope, *, dry_run: bool) -> bool:
        """
        Apply `soft_max_size` and `max_size` limits.

        This may involve moving files between stores.
        """
        ok = True

        # TODO: This doesn't handle the per-scope soft_max_size policy yet,
        # since that involves calculating a per-scope-and-store total size
        # and that's a bit involved.
        for store in (
            scope.file_stores.annotate_current_total_size()
            .filter(
                Q(soft_max_size__lt=F("current_total_size"))
                | Q(max_size__lt=F("current_total_size"))
            )
            .order_by("id")
        ):
            policies = scope.filestoreinscope_set.get(file_store=store)

            if policies.drain and dry_run:
                # In dry-run mode, we'll already have drained this store in
                # a previous step, so trying to do so again here just
                # produces confusing output.
                continue

            # Pre-compute the volume of data we need to drain, as otherwise
            # we can't provide a useful dry-run simulation.  This approach
            # isn't perfect because something else might still be adding
            # files to a store whose soft_max_size is exceeded, but the next
            # vacuum_storage run will catch up with that.
            current_total_size = (
                store.current_total_size  # type: ignore[attr-defined]
            )
            size_to_drain = max(
                [
                    current_total_size - limit
                    for limit in (store.soft_max_size, store.max_size)
                    if limit is not None
                ]
            )
            self.print_verbose(
                f"{scope}: Draining up to {size_to_drain} bytes from "
                f"{store.name}"
            )

            # It's not completely clear what we should do if a store's
            # soft_max_size/max_size is exceeded and the store is linked to
            # multiple scopes.  In principle we could try draining files
            # from any of its scopes.  For now we just take the simple
            # approach of draining files from each scope in order as long as
            # we need to, but in future we might instead decide to have some
            # kind of priority system.
            for file in (
                File.objects.filter(
                    artifact__workspace__scope=scope, filestore=store
                )
                .distinct()
                .order_by("artifact__created_at", F("size").desc(), "sha256")
            ):
                # Handle each file in a separate transaction; draining may
                # take some time, and we don't want to have to wait for the
                # entire store to be drained before files moved from it to
                # other stores are visible, or to roll back all the new
                # FileInStore rows because one move failed.
                with transaction.atomic():
                    # Because this is in a separate transaction from the
                    # query that found files to drain, it's possible that
                    # the file has since been deleted.  If so, we can ignore
                    # it.
                    try:
                        file.refresh_from_db()
                    except File.DoesNotExist:
                        continue

                    if self._drain_file(
                        scope=scope,
                        policies=policies,
                        store=store,
                        file=file,
                        dry_run=dry_run,
                    ):
                        size_to_drain -= file.size
                        if size_to_drain <= 0:
                            break
                    else:
                        ok = False

        return ok

    def _apply_storage_policies(self, *, dry_run: bool) -> bool:
        """
        Apply file storage policies.

        This may involve copying or moving files between stores in order to
        satisfy the `populate` and `drain` policies and the store-level
        `soft_max_size` and `max_size` limits.
        """
        ok = True

        for scope in Scope.objects.order_by("id"):
            if not self._apply_populate_policy(scope, dry_run=dry_run):
                ok = False

            if not self._apply_drain_policy(scope, dry_run=dry_run):
                ok = False

            if not self._apply_max_size_limits(scope, dry_run=dry_run):
                ok = False

        return ok

    def _delete_incomplete_old_artifacts(self, *, dry_run: bool) -> None:
        """Delete artifacts that are incomplete and created > 1 day ago."""
        self.print_verbose("Checking incomplete artifacts")

        one_day_ago = timezone.now() - dt.timedelta(days=1)

        to_delete = (
            Artifact.objects.filter(created_at__lt=one_day_ago)
            .annotate_complete()
            .filter(complete=False)
            # TODO: Incomplete artifacts ending up as collection items is
            # most likely a bug, but at the moment it does happen
            # occasionally.  Exclude such cases and let expiry take care of
            # them instead.
            .exclude(id__in=CollectionItem.objects.values("artifact"))
            .prefetch_related("fileinartifact_set")
        ).order_by("id")

        ArtifactRelation.objects.filter(
            Q(artifact__in=to_delete) | Q(target__in=to_delete)
        ).delete()

        for artifact in to_delete:
            for file_in_artifact in artifact.fileinartifact_set.all():
                if not dry_run:
                    if hasattr(file_in_artifact, "fileupload"):
                        file_in_artifact.fileupload.delete()

                    file_in_artifact.delete()

            artifact_id = artifact.id
            if not dry_run:
                artifact.delete()

            self.print_verbose(f"Deleted incomplete artifact {artifact_id}")

    def _delete_redundant_uploads(self, *, dry_run: bool) -> None:
        """
        Delete redundant ``FileUpload`` rows.

        These can happen when a client's attempt to upload a file is timed
        out by the load balancer, and the client retries.  In this case two
        parallel transactions may end up trying to upload the same file,
        because the server is not aware of the load balancer's timeout, and
        since both of them may add a ``FileUpload`` row and mark the same
        ``FileInArtifact`` row as complete without causing a transactional
        conflict, they may both succeed.

        ``FileUpload`` rows associated with complete ``FileInArtifact`` rows
        are pointless, so we can delete them after a short grace period to
        avoid conflicting with timed-out uploads that may still be in
        progress.
        """
        self.print_verbose("Checking redundant uploads")

        for upload in FileUpload.objects.filter(
            file_in_artifact__complete=True,
            last_activity_at__lt=timezone.now() - dt.timedelta(hours=1),
        ):
            upload_id = upload.id
            if not dry_run:
                upload.delete()

            self.print_verbose(f"Deleted incomplete upload {upload_id}")

    def _delete_orphan_files_from_upload_directory(
        self, *, dry_run: bool
    ) -> None:
        """Delete files from the upload directory that are not referenced."""
        self.print_verbose("Checking orphan files in upload directory")

        paths_in_disk = self._list_directory(
            settings.DEBUSINE_UPLOAD_DIRECTORY,
            filters=[self._path_older_than_one_day],
        )

        paths_in_db = set()
        for file_upload in FileUpload.objects.all():
            paths_in_db.add(file_upload.absolute_file_path())

        files_to_remove = paths_in_disk - paths_in_db

        for file_to_remove in files_to_remove:
            if not dry_run:
                file_to_remove.unlink(missing_ok=True)

            self.print_verbose(f"Deleted {file_to_remove}")

    def _report_missing_files_stores_referenced_by_the_db(self) -> bool:
        """
        Report DB referencing files that do not exist in storage.

        It cannot be fixed automatically.
        """
        self.print_verbose("Checking missing files from stores")
        ok = True

        for store in FileStore.objects.order_by("id"):
            # TODO: We should do this in a way that doesn't require holding
            # both sets in memory.
            entries_in_db = self._entries_from_files_in_file_store_db(store)
            entries_in_storage = set(store.get_backend_object().list_entries())
            orphan_entries = entries_in_db - entries_in_storage

            for orphan_entry in orphan_entries:
                self.stderr.write(
                    f"File in FileStore {store.name} but not in storage: "
                    f"{orphan_entry}"
                )

            ok = ok and (len(orphan_entries) == 0)

        return ok

    @staticmethod
    def _entries_from_files_in_file_store_db(
        store: FileStore,
    ) -> set[FileBackendEntryInterface[Any, Any]]:
        store_backend = store.get_backend_object()
        # TODO: This should probably also filter to files that are in any of
        # the scopes configured to use this store.  After `debusine-admin
        # scope manage --remove-file-store`, it's possible for the store to
        # contain files that are no longer in any of its scopes, and those
        # should be cleaned up.
        return {
            store_backend.get_entry(file_in_store.file)
            for file_in_store in FileInStore.objects.filter(
                store=store
            ).select_related("file")
        }

    def _report_missing_files_referenced_by_db_file_upload(self) -> bool:
        """Report DB referencing upload files that do not exist in disk."""
        self.print_verbose("Checking missing files from upload")

        files_in_disk = self._list_directory(settings.DEBUSINE_UPLOAD_DIRECTORY)

        files_in_db = {
            file_upload.absolute_file_path()
            # Skip uploads that are likely to be deleted by
            # _delete_redundant_uploads in the next run.
            for file_upload in FileUpload.objects.exclude(
                file_in_artifact__complete=True
            )
        }

        orphan_files = files_in_db - files_in_disk

        for orphan_file in orphan_files:
            self.stderr.write(
                f"File in FileUpload but not on disk: {orphan_file}"
            )

        return len(orphan_files) == 0

    @staticmethod
    def _list_directory(
        directory: Path | str,
        *,
        filters: list[Callable[[Path], bool]] | None = None,
    ) -> set[Path]:
        if filters is None:
            filters = []

        # Pathlib.glob documentation: "Using the “**” pattern in large
        # directory trees may consume an inordinate amount of time."
        # Consider changing this with something else (e.g. os.walk; or
        # yielding paths and checking the DB per each file/set of files if
        # this becomes a problem).
        paths = Path(directory).glob("**/*")

        filtered_paths = set()

        for path in paths:
            if all(f(path) for f in filters):
                filtered_paths.add(path)

        return filtered_paths

    def _roll_up_file_store_journal(self, *, dry_run: bool) -> None:
        """Roll up entries from FileStoreJournal into FileStore.total_size."""
        for store in FileStore.objects.annotate_current_total_size():
            current_total_size = (
                store.current_total_size  # type: ignore[attr-defined]
            )
            if current_total_size != store.total_size and not dry_run:
                store.total_size = current_total_size
                store.save()
                store.journal_entries.filter(
                    pk__lte=store.journal_entries.aggregate(max_pk=Max("pk"))[
                        "max_pk"
                    ]
                ).delete()



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for the DB-aware mock field resolver."""

from debusine.artifacts.models import CollectionCategory
from debusine.db.tasks.mock_resolver import MockFieldResolver
from debusine.db.tasks.models import DBArtifactInfo, DBInputCollectionSingle
from debusine.test.django import TestCase


class MockFieldResolverTests(TestCase):
    """Tests for :py:class:`MockFieldResolver`."""

    def test_set_artifact(self) -> None:
        resolver = MockFieldResolver()
        artifact = self.playground.create_minimal_binary_package_artifact()
        resolver.set_artifact("test", artifact)
        self.assertEqual(
            resolver.field_values["test"],
            DBArtifactInfo.from_artifact(
                artifact, lookup=artifact.id, label="test"
            ),
        )

    def test_set_collection(self) -> None:
        resolver = MockFieldResolver()
        collection = self.playground.create_collection(
            name="sid", category=CollectionCategory.SUITE
        )
        resolver.set_collection("test", collection)
        self.assertEqual(
            resolver.field_values["test"],
            DBInputCollectionSingle.from_collection(
                collection, lookup=collection.id, label="test"
            ),
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Task to upload Debian packages to an upload queue."""

import datetime as dt
import shutil
import tempfile
from ftplib import FTP
from pathlib import Path
from typing import override

import tenacity
from fabric import Connection

from debusine.artifacts.models import ArtifactCategory, WorkRequestResults
from debusine.db.tasks.inputs import ArtifactInput
from debusine.server.tasks import BaseServerTask
from debusine.server.tasks.models import (
    PackageUploadData,
    PackageUploadDynamicData,
    PackageUploadTarget,
)

# At some point it might be useful to make this configurable, but hardcoding
# it will do for now.
_TIMEOUT = dt.timedelta(seconds=15)


class PackageUpload(
    BaseServerTask[PackageUploadData, PackageUploadDynamicData]
):
    """Task that uploads Debian packages to an upload queue."""

    TASK_VERSION = 1
    upload = ArtifactInput(field="input.upload")

    @override
    def compute_dynamic_data(self) -> PackageUploadDynamicData:
        """Resolve artifact lookups for this task."""
        dynamic_data = PackageUploadDynamicData(
            input_upload_id=self.upload.artifact_id
        )
        self.save_field_values(dynamic_data)
        return dynamic_data

    def fetch_upload(self, destination: Path) -> list[Path] | None:
        """Download the required artifacts."""
        upload = self.upload.artifact
        if upload.category != ArtifactCategory.UPLOAD:
            self.append_to_log_file(
                "fetch_upload.log",
                [
                    f"Expected input.upload of category "
                    f"{ArtifactCategory.UPLOAD}; got {upload.category}"
                ],
            )
            return None

        upload_paths: list[Path] = []
        for file_in_artifact in upload.fileinartifact_set.select_related(
            "file"
        ):
            file_path = destination / file_in_artifact.path
            # Should be checked by LocalArtifact, but let's make sure.
            if not file_path.resolve().is_relative_to(destination):
                raise AssertionError(
                    f"{file_in_artifact.path} escapes directory"
                )
            file = file_in_artifact.file
            file_backend = upload.workspace.scope.download_file_backend(file)
            with (
                file_backend.get_stream(file) as infile,
                file_path.open(mode="wb") as outfile,
            ):
                shutil.copyfileobj(infile, outfile)
            upload_paths.append(file_path)

        return sorted(
            upload_paths,
            # Some upload queues use the appearance of the .changes file as
            # an indication that the upload is complete, so sort it to the
            # end.
            key=lambda path: (path.name.endswith(".changes"), path),
        )

    def _make_target(self) -> PackageUploadTarget:
        """Make the full upload target URL."""
        full_target = str(self.data.target)
        if self.data.delayed_days is not None:
            full_target = (
                f"{full_target.rstrip('/')}/"
                f"DELAYED/{self.data.delayed_days}-day"
            )
        return PackageUploadTarget(full_target)

    @tenacity.retry(
        stop=tenacity.stop_after_attempt(3),
        retry=(
            tenacity.retry_if_exception_type(OSError)
            | tenacity.retry_if_exception_type(TimeoutError)
        ),
        reraise=True,
    )
    def _upload_ftp(
        self, target: PackageUploadTarget, upload_paths: list[Path]
    ) -> None:
        """Make an upload using FTP."""
        assert target.host is not None

        try:
            with FTP(timeout=_TIMEOUT.total_seconds()) as ftp:
                ftp.connect(host=target.host, port=int(target.port or 0))
                ftp.login(
                    user=target.username or "", passwd=target.password or ""
                )
                assert target.path is not None
                ftp.cwd(target.path)
                for path in upload_paths:
                    self.append_to_log_file(
                        "package-upload.log", [f"Uploading {path.name}"]
                    )
                    with path.open(mode="rb") as f:
                        ftp.storbinary(f"STOR {path.name}", f)
        except Exception as e:
            self.append_to_log_file(
                "package-upload.log", [f"Upload failed: {e}"]
            )
            raise

    @tenacity.retry(
        stop=tenacity.stop_after_attempt(3),
        retry=(
            tenacity.retry_if_exception_type(OSError)
            | tenacity.retry_if_exception_type(TimeoutError)
        ),
        reraise=True,
    )
    def _upload_sftp(
        self, target: PackageUploadTarget, upload_paths: list[Path]
    ) -> None:
        """Make an upload using SFTP."""
        assert target.host is not None

        try:
            with Connection(
                host=target.host,
                user=target.username,
                port=None if target.port is None else int(target.port),
                connect_timeout=_TIMEOUT.total_seconds(),
            ) as connection:
                for path in upload_paths:
                    self.append_to_log_file(
                        "package-upload.log", [f"Uploading {path.name}"]
                    )
                    connection.put(
                        path, str(Path(target.path or "", path.name))
                    )
        except Exception as e:
            self.append_to_log_file(
                "package-upload.log", [f"Upload failed: {e}"]
            )
            raise

    @override
    def _execute(self) -> WorkRequestResults:
        """Execute the task."""
        with tempfile.TemporaryDirectory(
            prefix="debusine-package-upload-"
        ) as temp_dir:
            temp_path = Path(temp_dir)
            upload_paths = self.fetch_upload(temp_path)
            if upload_paths is None:
                return WorkRequestResults.FAILURE
            target = self._make_target()

            self.append_to_log_file(
                "package-upload.log", [f"Uploading to {target}"]
            )
            match target.scheme:
                case "ftp":
                    self._upload_ftp(target, upload_paths)
                case "sftp":
                    self._upload_sftp(target, upload_paths)
                case _ as unreachable:
                    raise AssertionError(
                        f"Unexpected URL scheme: {unreachable}"
                    )
            self.append_to_log_file("package-upload.log", ["Upload succeeded"])
            return WorkRequestResults.SUCCESS

    @override
    def get_label(self) -> str:
        """Return the task label."""
        return f"upload to {self._make_target()}"



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for the file widgets."""

import json
import os.path
import textwrap
import unittest
from typing import ClassVar, NamedTuple, cast, override
from unittest import mock

from django.http.request import HttpHeaders
from django.http.response import HttpResponseBase, HttpResponseRedirectBase
from django.template import engines
from django.test import override_settings
from django.utils.datastructures import CaseInsensitiveMapping
from django.utils.html import escape
from django.utils.safestring import SafeString, mark_safe
from pydantic import ValidationError
from rest_framework import status

from debusine.artifacts.models import ArtifactCategory
from debusine.db.models import FileInArtifact
from debusine.server.views import ProblemResponse
from debusine.test.django import TestCase
from debusine.web.views import files
from debusine.web.views.files import (
    FileDownloadMixin,
    FileUI,
    FileWidget,
    LinenoHtmlFormatter,
    LogFileWidget,
    MAX_FILE_SIZE,
    PathMixin,
    add_line_number,
)
from debusine.web.views.tests.utils import ViewTestMixin


class FileUITests(TestCase):
    """Tests for the FileUI class."""

    def assertRenders(
        self, file_ui: FileUI, file_in_artifact: FileInArtifact
    ) -> None:
        """Test that the given widget renders without errors."""
        template = engines["django"].from_string(
            "{% load debusine %}{% widget widget %}"
        )
        setattr(template, "engine", engines["django"])
        widget = file_ui.widget_class(file_in_artifact, file_ui=file_ui)
        # Set DEBUG=True to have render raise exceptions instead of logging
        # them
        with override_settings(DEBUG=True):
            template.render({"widget": widget})

    def test_from_file_in_artifact_explicit_content_type(self) -> None:
        """Test from_file_in_artifact with explicit content-types."""

        class SubTest(NamedTuple):
            """Define a subtest."""

            filename: str
            db_content_type: str
            widget_class: type[FileWidget]
            rendered_content_type: str
            size: int = 1024

        sub_tests = [
            SubTest(
                "build.changes",
                "text/plain; charset=utf-8",
                files.TextFileWidget,
                "text/plain; charset=utf-8",
            ),
            SubTest(
                "file.log",
                "text/plain; charset=utf-8",
                files.TextFileWidget,
                "text/plain; charset=utf-8",
            ),
            SubTest(
                "file.txt",
                "text/plain; charset=utf-8",
                files.TextFileWidget,
                "text/plain; charset=utf-8",
            ),
            SubTest(
                "hello.build",
                "text/plain; charset=utf-8",
                files.TextFileWidget,
                "text/plain; charset=utf-8",
            ),
            SubTest(
                "hello.buildinfo",
                "text/plain; charset=utf-8",
                files.TextFileWidget,
                "text/plain; charset=utf-8",
            ),
            SubTest(
                "file.sources",
                "text/plain; charset=utf-8",
                files.TextFileWidget,
                "text/plain; charset=utf-8",
            ),
            SubTest(
                "readme.md",
                "text/markdown; charset=utf-8",
                files.TextFileWidget,
                "text/markdown; charset=utf-8",
            ),
            SubTest(
                "log",
                "text/plain; charset=utf-8",
                files.TextFileWidget,
                "text/plain; charset=utf-8",
            ),
            SubTest(
                "file.js", "text/javascript", files.TextFileWidget, "text/plain"
            ),
            SubTest(
                "output.json",
                "application/json; charset=us-ascii",
                files.TextFileWidget,
                "application/json; charset=us-ascii",
            ),
            SubTest(
                "empty",
                "inode/x-empty; charset=binary",
                files.TextFileWidget,
                "text/plain",
            ),
            SubTest(
                "Release.gpg",
                "application/pgp-signature; charset=utf-8",
                files.TextFileWidget,
                "application/pgp-signature; charset=utf-8",
            ),
            SubTest(
                "InRelease",
                "text/PGP; charset=utf-8",
                files.TextFileWidget,
                "text/PGP; charset=utf-8",
            ),
            SubTest(
                "a.out",
                "application/octet-stream",
                files.BinaryFileWidget,
                "application/octet-stream",
            ),
            SubTest(
                "big.bin",
                "application/octet-stream",
                files.TooBigFileWidget,
                "application/octet-stream",
                2**32,
            ),
            SubTest(
                "image.png",
                "application/png",
                files.BinaryFileWidget,
                "application/octet-stream",
            ),
        ]

        for sub_test in sub_tests:
            with self.subTest(sub_test.filename):
                artifact, _ = self.playground.create_artifact(
                    paths={sub_test.filename: b"test"}, create_files=True
                )
                file_in_artifact = FileInArtifact.objects.get(artifact=artifact)
                file_in_artifact.content_type = sub_test.db_content_type
                file_in_artifact.file.size = sub_test.size
                fileui = FileUI.from_file_in_artifact(file_in_artifact)
                self.assertEqual(
                    fileui.content_type, sub_test.rendered_content_type
                )
                self.assertEqual(fileui.widget_class, sub_test.widget_class)
                self.assertRenders(fileui, file_in_artifact)

    def test_from_file_in_artifact_guess_from_file_name(self) -> None:
        """Test from_file_in_artifact, guessing based on the file name."""

        class SubTest(NamedTuple):
            """Define a subtest."""

            filename: str
            widget_class: type[FileWidget]
            content_type: str
            size: int = 1024

        sub_tests = [
            SubTest(
                "build.changes",
                files.TextFileWidget,
                "text/plain; charset=utf-8",
            ),
            SubTest(
                "file.log", files.TextFileWidget, "text/plain; charset=utf-8"
            ),
            SubTest(
                "file.txt", files.TextFileWidget, "text/plain; charset=utf-8"
            ),
            SubTest(
                "hello.build", files.TextFileWidget, "text/plain; charset=utf-8"
            ),
            SubTest(
                "hello.buildinfo",
                files.TextFileWidget,
                "text/plain; charset=utf-8",
            ),
            SubTest(
                "file.sources",
                files.TextFileWidget,
                "text/plain; charset=utf-8",
            ),
            SubTest(
                "readme.md",
                files.TextFileWidget,
                "text/markdown; charset=utf-8",
            ),
            SubTest(
                "a.out", files.BinaryFileWidget, "application/octet-stream"
            ),
            SubTest(
                "big.bin",
                files.TooBigFileWidget,
                "application/octet-stream",
                2**32,
            ),
        ]

        for sub_test in sub_tests:
            with self.subTest(sub_test.filename):
                artifact, _ = self.playground.create_artifact(
                    paths={sub_test.filename: b"test"}, create_files=True
                )
                file_in_artifact = FileInArtifact.objects.get(artifact=artifact)
                file_in_artifact.file.size = sub_test.size
                fileui = FileUI.from_file_in_artifact(file_in_artifact)
                self.assertEqual(fileui.content_type, sub_test.content_type)
                self.assertEqual(fileui.widget_class, sub_test.widget_class)
                self.assertRenders(fileui, file_in_artifact)

    def test_from_file_in_artifact_autopkgtest_log(self) -> None:
        artifact, _ = self.playground.create_artifact(
            paths=["log"],
            create_files=True,
            category=ArtifactCategory.AUTOPKGTEST,
        )
        file_in_artifact = artifact.fileinartifact_set.get(path="log")

        file_ui = FileUI.from_file_in_artifact(file_in_artifact)

        self.assertEqual(
            file_ui,
            FileUI(
                content_type="text/plain; charset=utf-8",
                widget_class=LogFileWidget,
                pygments_lexer=None,
            ),
        )

    def test_compressed(self) -> None:
        """Test from_file_in_artifact with a compressed file."""
        artifact, _ = self.playground.create_artifact(
            paths=["file.md.gz"], create_files=True
        )
        fileui = FileUI.from_file_in_artifact(
            artifact.fileinartifact_set.get(path="file.md.gz")
        )

        self.assertEqual(fileui.content_type, "application/gzip")
        self.assertEqual(fileui.widget_class, files.BinaryFileWidget)


class PathMixinTests(ViewTestMixin, unittest.TestCase):
    """Tests for the ArtifactDetailView class."""

    def test_normalize_path(self) -> None:
        """Test PathMixin.normalize_path."""
        f = PathMixin.normalize_path
        self.assertEqual(f(""), "/")
        self.assertEqual(f("/"), "/")
        self.assertEqual(f("."), "/")
        self.assertEqual(f(".."), "/")
        self.assertEqual(f("../"), "/")
        self.assertEqual(f("../../.././../../"), "/")
        self.assertEqual(f("src/"), "/src/")
        self.assertEqual(f("src/.."), "/")
        self.assertEqual(f("/a/b/../c/./d//e/f/../g/"), "/a/c/d/e/g/")

    def test_path(self) -> None:
        """Test PathMixin.path."""
        view = self.instantiate_view_class(PathMixin, "/")
        self.assertEqual(view.path, "")

        view = self.instantiate_view_class(
            PathMixin, "/", path="/foo/bar/../baz"
        )
        self.assertEqual(view.path, "foo/baz/")


class FileDownloadMixinTests(ViewTestMixin, TestCase):
    """Test FileDownloadMixin."""

    playground_memory_file_store = False

    contents: ClassVar[dict[str, bytes]]
    empty: ClassVar[FileInArtifact]
    file: ClassVar[FileInArtifact]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.contents = {
            "empty.bin": b"",
            "file.md": bytes(range(256)),
        }
        artifact, _ = cls.playground.create_artifact(
            paths=cls.contents,
            create_files=True,
        )
        cls.empty = artifact.fileinartifact_set.get(path="empty.bin")
        cls.file = artifact.fileinartifact_set.get(path="file.md")

    def get_stream_response(
        self,
        file_in_artifact: FileInArtifact,
        range_header: tuple[int | None, int | None] | str | None = None,
        download: bool = True,
    ) -> HttpResponseBase:
        """Instantiate the view and get a streaming file response."""
        request = self.make_request("/")
        headers = {}
        match range_header:
            case str():
                headers["Range"] = range_header
            case [range_start, range_end]:
                if range_start is None:
                    range_start = ""
                if range_end is None:
                    range_end = ""
                headers["Range"] = f"bytes={range_start}-{range_end}"
        if headers:
            request.headers = cast(
                HttpHeaders,
                CaseInsensitiveMapping({**request.headers, **headers}),
            )

        ui_info = FileUI.from_file_in_artifact(file_in_artifact)
        view = self.instantiate_view_class(FileDownloadMixin, request)
        response = view.stream_file(file_in_artifact, ui_info, download)
        if isinstance(response, ProblemResponse):
            # This is needed by assertResponseProblem
            setattr(
                response,
                "json",
                lambda: json.loads(response.content.decode(response.charset)),
            )

        return response

    def assertFileResponse(
        self,
        response: HttpResponseBase,
        file_in_artifact: FileInArtifact,
        range_start: int = 0,
        range_end: int | None = None,
        disposition: str = "attachment",
    ) -> None:
        """Assert that response has the expected headers and content."""
        if range_start == 0 and range_end is None:
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        else:
            self.assertEqual(
                response.status_code, status.HTTP_206_PARTIAL_CONTENT
            )

        file_contents = self.contents[file_in_artifact.path]
        if range_end is None:
            range_end = len(file_contents) - 1
        expected_contents = file_contents[range_start : range_end + 1]

        headers = response.headers
        self.assertEqual(headers["Accept-Ranges"], "bytes")
        self.assertEqual(headers["Content-Length"], str(len(expected_contents)))

        if len(expected_contents) > 0:
            self.assertEqual(
                headers["Content-Range"],
                f"bytes {range_start}-{range_end}/{len(file_contents)}",
            )

        filename = os.path.basename(file_in_artifact.path)
        self.assertEqual(
            headers["Content-Disposition"],
            f'{disposition}; filename="{filename}"',
        )

        streaming_content = getattr(response, "streaming_content", None)
        assert streaming_content is not None
        self.assertEqual(b"".join(streaming_content), expected_contents)

    def test_get_file(self) -> None:
        """Get return the file."""
        response = self.get_stream_response(self.file)
        self.assertFileResponse(response, self.file)
        self.assertEqual(
            response.headers["content-type"], "text/markdown; charset=utf-8"
        )

    def test_get_file_inline(self) -> None:
        """Get return the file."""
        response = self.get_stream_response(self.file, download=False)
        self.assertFileResponse(response, self.file, disposition="inline")
        self.assertEqual(
            response.headers["content-type"], "text/markdown; charset=utf-8"
        )

    def test_get_empty_file(self) -> None:
        """Test empty downloadable file (which mmap doesn't support)."""
        response = self.get_stream_response(self.empty)
        self.assertFileResponse(response, self.empty)
        self.assertEqual(
            response.headers["content-type"],
            "application/octet-stream",
        )

    def test_get_incomplete_file(self) -> None:
        """Get returns 404 for an incomplete file."""
        self.file.complete = False
        self.file.save()
        response = self.get_stream_response(self.file)
        self.assertResponseProblem(
            response,
            "Cannot download incomplete file",
            status_code=status.HTTP_404_NOT_FOUND,
        )

    def test_get_file_range(self) -> None:
        """Get return part of the file (based on Range header)."""
        start, end = 10, 20
        response = self.get_stream_response(
            self.file, range_header=(start, end)
        )
        self.assertFileResponse(response, self.file, start, end)

    def test_get_file_content_range_to_end_of_file(self) -> None:
        """Server returns a file from a position to the end."""
        start, end = 5, len(self.contents["file.md"]) - 1
        response = self.get_stream_response(
            self.file, range_header=(start, end)
        )
        self.assertFileResponse(response, self.file, start, end)

    def test_get_file_range_no_end_position(self) -> None:
        """Server handles ``Range`` with no end position."""
        start, end = 5, len(self.contents["file.md"]) - 1
        response = self.get_stream_response(
            self.file, range_header=(start, None)
        )
        self.assertFileResponse(response, self.file, start, end)

    def test_get_file_range_suffix(self) -> None:
        """Server handles ``Range`` with a suffix byte-range request."""
        suffix_length = 10
        end = len(self.contents["file.md"]) - 1
        start = end - suffix_length + 1
        response = self.get_stream_response(
            self.file, range_header=(None, suffix_length)
        )
        self.assertFileResponse(response, self.file, start, end)

    def test_get_file_content_range_invalid(self) -> None:
        """Get return an error: Range header was invalid."""
        invalid_range_header = "invalid-range"
        response = self.get_stream_response(
            self.file, range_header=invalid_range_header
        )
        self.assertResponseProblem(
            response, f'Invalid Range header: "{invalid_range_header}"'
        )

    def test_get_file_content_range_no_positions(self) -> None:
        """Get return an error: Range header contains neither start nor end."""
        invalid_range_header = "bytes=-"
        response = self.get_stream_response(
            self.file, range_header=invalid_range_header
        )
        self.assertResponseProblem(
            response, f'Invalid Range header: "{invalid_range_header}"'
        )

    def test_get_file_range_start_greater_file_size(self) -> None:
        """Get return 400: client requested an invalid start position."""
        file_size = len(self.contents["file.md"])
        start, end = file_size + 10, file_size + 20
        response = self.get_stream_response(
            self.file, range_header=(start, end)
        )
        self.assertResponseProblem(
            response,
            f"Invalid Content-Range start: {start}. File size: {file_size}",
        )

    def test_get_file_range_end_is_file_size(self) -> None:
        """Get return 400: client requested and invalid end position."""
        end = len(self.contents["file.md"])
        response = self.get_stream_response(self.file, range_header=(0, end))
        self.assertResponseProblem(
            response,
            f"Invalid Content-Range end: {end}. File size: {end}",
        )

    def test_get_file_range_end_greater_file_size(self) -> None:
        """Get return 400: client requested an invalid end position."""
        file_size = len(self.contents["file.md"])
        end = file_size + 10
        response = self.get_stream_response(self.file, range_header=(0, end))
        self.assertResponseProblem(
            response,
            f"Invalid Content-Range end: {end}. File size: {file_size}",
        )

    def test_get_file_url_redirect(self) -> None:
        """
        Get file response: redirect if get_url for the file is available.

        This would happen if the file is stored in a FileStore supporting
        get_url (e.g. an object storage) instead of being served from the
        server's file system.
        """
        destination_url = "https://some-backend.net/file?token=asdf"

        with mock.patch(
            "debusine.server.file_backend.local.LocalFileBackendEntry.get_url",
            autospec=True,
            return_value=destination_url,
        ) as mock_get_url:
            response = self.get_stream_response(self.file)
            self.assertEqual(response.status_code, status.HTTP_302_FOUND)
            assert isinstance(response, HttpResponseRedirectBase)
            self.assertEqual(response.url, destination_url)
            ui_info = FileUI.from_file_in_artifact(self.file)
            mock_get_url.assert_called_once_with(
                mock.ANY, content_type=ui_info.content_type
            )


class FileWidgetTests(ViewTestMixin, TestCase):
    """Test FileView."""

    contents: ClassVar[dict[str, bytes]]
    binary: ClassVar[FileInArtifact]
    dsc: ClassVar[FileInArtifact]
    empty: ClassVar[FileInArtifact]
    large: ClassVar[FileInArtifact]
    text: ClassVar[FileInArtifact]
    json: ClassVar[FileInArtifact]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.contents = {
            "file.md": b"# Title\nText",
            "file.dsc": b"Source: hello",
            "empty.bin": b"",
            "file.bin": bytes(range(256)),
            "largefile.bin": b"large",
            "output.json": b'{"name": "John"}',
        }
        artifact, _ = cls.playground.create_artifact(
            paths=cls.contents,
            create_files=True,
        )
        cls.empty = artifact.fileinartifact_set.get(path="empty.bin")
        cls.dsc = artifact.fileinartifact_set.get(path="file.dsc")
        cls.text = artifact.fileinartifact_set.get(path="file.md")
        cls.json = artifact.fileinartifact_set.get(path="output.json")
        cls.json.content_type = "application/json; charset=us-ascii"
        cls.json.save()
        cls.binary = artifact.fileinartifact_set.get(path="file.bin")
        cls.large = artifact.fileinartifact_set.get(path="largefile.bin")
        cls.large.file.size = 2**32
        cls.large.file.save()

    def test_context_data(self) -> None:
        """Test context for file types that are supported."""
        for file, expected_content in [
            (self.text, "# Title"),
            (self.json, "John"),
            (self.dsc, '<span class="k">Source</span>'),
        ]:
            with self.subTest(file=file.path):
                widget = FileWidget.create(file)
                context = widget.get_context_data()

                file_content = context.pop("file_content")
                self.assertIn(expected_content, file_content)

                self.assertEqual(
                    context,
                    {
                        "file_in_artifact": file,
                        "file_lineno_width": 2,
                        "file_ui": FileUI.from_file_in_artifact(file),
                        "formatted_content_div_id": "formatted-content",
                        "widget": widget,
                    },
                )

    def test_context_data_text_incomplete(self) -> None:
        """Test context for incomplete text files."""
        file = self.text
        file.complete = False
        file.save()
        widget = FileWidget.create(file)
        context = widget.get_context_data()

        self.assertNotIn("file_content", context)

        self.assertEqual(
            context,
            {
                "file_in_artifact": file,
                "file_ui": FileUI.from_file_in_artifact(file),
                "formatted_content_div_id": "formatted-content",
                "widget": widget,
            },
        )

    def test_context_data_empty(self) -> None:
        """Test context for empty binary files."""
        file = self.empty
        widget = FileWidget.create(file)
        context = widget.get_context_data()
        self.assertEqual(
            context,
            {
                "file_in_artifact": file,
                "file_ui": FileUI.from_file_in_artifact(file),
                "widget": widget,
            },
        )

    def test_text_file_line_numbers(self) -> None:
        """Test a text file has the line numbers."""
        file = self.text
        context = FileWidget.create(file).get_context_data()
        self.assertHTMLEqual(
            context["file_content"],
            textwrap.dedent(
                """\
                <div class="formatted-content">
                <a id="L1" href="#L1">1</a><i># Title\n</i>
                <a id="L2" href="#L2">2</a><i>Text\n</i>
                </div>\n
                """
            ),
        )

    def test_context_data_binary(self) -> None:
        """Test context for binary files."""
        file = self.binary
        widget = FileWidget.create(file)
        context = widget.get_context_data()
        self.assertEqual(
            context,
            {
                "file_in_artifact": file,
                "file_ui": FileUI.from_file_in_artifact(file),
                "widget": widget,
            },
        )

    def test_context_data_large(self) -> None:
        """Test context for files too large."""
        file = self.large
        widget = FileWidget.create(file)
        context = widget.get_context_data()
        self.assertEqual(
            context,
            {
                "file_in_artifact": file,
                "file_ui": FileUI.from_file_in_artifact(file),
                "file_max_size": MAX_FILE_SIZE,
                "widget": widget,
            },
        )

    def test_read_file_from_fileinartifact(self) -> None:
        file = self.binary
        widget = FileWidget.create(file)

        self.assertEqual(widget.read_file(), self.contents["file.bin"])

        # File contents is cached
        self.assertEqual(widget.file_contents, self.contents["file.bin"])

    def test_read_file_from_contents(self) -> None:
        file = self.binary
        widget = FileWidget.create(file, file_contents=b"already-read-contents")

        # File contents is used instead of reading self.binary
        self.assertEqual(widget.read_file(), b"already-read-contents")

    def test_initial_contents_discarded(self) -> None:
        file = self.binary
        big_contents = b"x" * (MAX_FILE_SIZE + 1)

        widget = FileWidget.create(file, file_contents=big_contents)

        # Contents is too big: discarded
        self.assertIsNone(widget.file_contents)

        # Contents was discarded: read_file() tries to read and succeeds
        self.assertEqual(widget.read_file(), self.contents["file.bin"])


class LogFileWidgetTests(ViewTestMixin, TestCase):
    """Tests for LogFileWidget."""

    contents: ClassVar[dict[str, bytes]]
    file: ClassVar[FileInArtifact]

    before_contents = (
        '<div class="formatted-content"><pre class="m-0 p-0"><span></span>'
    )
    after_contents = '</pre></div>'

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.contents = {
            "log": textwrap.dedent(
                """\
            first line
            second line\rprogress\rprogress\rfinished
            third line
            fourth line"""
            ).encode(),
            "output.json": json.dumps({"key": "value"}).encode(),
        }
        artifact, _ = cls.playground.create_artifact(
            paths=cls.contents,
            create_files=True,
        )
        cls.file = artifact.fileinartifact_set.get(path="log")

    @staticmethod
    def render(widget: LogFileWidget) -> str:
        template = engines["django"].from_string(
            "{% load debusine %}{% widget widget %}"
        )
        setattr(template, "engine", engines["django"])
        # Set DEBUG=True to have render raise exceptions instead of logging
        # them
        with override_settings(DEBUG=True):
            return template.render({"widget": widget})

    def test_get_context(self) -> None:
        widget = LogFileWidget(
            self.file,
            sections=[
                LogFileWidget.Section(
                    title="Start",
                    start_line=1,
                    end_line=2,
                    expanded=True,
                ),
                LogFileWidget.Section(
                    title="End",
                    start_line=3,
                    end_line=4,
                    expanded=True,
                ),
            ],
        )

        context = widget.get_context_data()

        # Sections as expected
        self.assertEqual(
            context["sections"],
            [
                LogFileWidget.RenderedSection(
                    expanded=True,
                    content=add_line_numbers(
                        start=1,
                        lines=[
                            "first line",
                            "second line^Mprogress^Mprogress^Mfinished",
                        ],
                    ),
                    start_line=1,
                    title="Start",
                ),
                LogFileWidget.RenderedSection(
                    expanded=True,
                    content=add_line_numbers(
                        start=3, lines=["third line", "fourth line"]
                    ),
                    start_line=3,
                    title="End",
                ),
            ],
        )

        self.assertIsInstance(context["file_ui"], FileUI)
        self.assertEqual(context["file_in_artifact"], self.file)
        self.assertEqual(
            context["formatted_content_div_id"], "formatted-content"
        )
        self.assertCountEqual(
            context.keys(),
            {
                "formatted_content_div_id",
                "file_in_artifact",
                "file_lineno_width",
                "file_ui",
                "sections",
                "suffix",
                "widget",
            },
        )

    def test_get_context_without_sections(self) -> None:
        widget = LogFileWidget(self.file, sections=[])

        context = widget.get_context_data()
        self.assertEqual(
            context["sections"],
            [
                LogFileWidget.RenderedSection(
                    start_line=1,
                    content=add_line_numbers(
                        start=1,
                        lines=[
                            "first line",
                            "second line^Mprogress^Mprogress^Mfinished",
                            "third line",
                            "fourth line",
                        ],
                    ),
                    expanded=True,
                    title=None,
                )
            ],
        )

    def test_get_context_with_sections_add_beginning_and_end(self) -> None:
        widget = LogFileWidget(
            self.file,
            file_tag="tag",
            sections=[
                LogFileWidget.Section(
                    title="Middle",
                    start_line=2,
                    end_line=3,
                    expanded=True,
                ),
            ],
        )

        context = widget.get_context_data()

        # Sections as expected
        self.assertEqual(
            context["sections"],
            [
                LogFileWidget.RenderedSection(
                    content=add_line_numbers(
                        start=1, suffix="-tag", lines=["first line"]
                    ),
                    expanded=True,
                    start_line=1,
                    title=None,
                ),
                LogFileWidget.RenderedSection(
                    content=add_line_numbers(
                        start=2,
                        suffix="-tag",
                        lines=[
                            "second line^Mprogress^Mprogress^Mfinished",
                            "third line",
                        ],
                    ),
                    expanded=True,
                    start_line=2,
                    title="Middle",
                ),
                LogFileWidget.RenderedSection(
                    content=add_line_numbers(
                        start=4, suffix="-tag", lines=["fourth line"]
                    ),
                    expanded=True,
                    start_line=4,
                    title=None,
                ),
            ],
        )
        self.assertEqual(context["suffix"], "-tag")
        self.assertIsInstance(context["file_ui"], FileUI)
        self.assertEqual(context["file_in_artifact"], self.file)
        self.assertEqual(
            context["formatted_content_div_id"], "formatted-content-tag"
        )
        self.assertCountEqual(
            context.keys(),
            {
                "formatted_content_div_id",
                "file_in_artifact",
                "file_lineno_width",
                "file_ui",
                "sections",
                "suffix",
                "widget",
            },
        )

    def test_contents_displayed_sections(self) -> None:
        rendered = self.render(
            LogFileWidget(
                self.file,
                sections=[
                    LogFileWidget.Section(
                        title="A title",
                        start_line=2,
                        end_line=3,
                        expanded=False,
                    )
                ],
            )
        )

        log_contents = (
            self.contents["log"].decode().replace("\r", "^M").split("\n")
        )

        tree = self.assertHTMLValid(rendered)

        file_contents = self.assertHasElement(
            tree, '//div[@id="formatted-content"]'
        )

        card_body = self.assertHasElement(
            file_contents, './/div[contains(@class, "card-body")]'
        )

        # Assert two untitled sections (lines that are not collapsible)
        first_untitled_section = self.assertHasElement(
            card_body, './pre[@class="formatted-content"][1]'
        )
        self.assertEqual(len(first_untitled_section.xpath("./i")), 1)
        self.assertEqual(
            first_untitled_section.xpath("./i")[0].text, log_contents[0]
        )

        second_untitled_section = self.assertHasElement(
            card_body, './pre[@class="formatted-content"][2]'
        )
        self.assertEqual(len(second_untitled_section.xpath("./i")), 1)
        self.assertEqual(
            second_untitled_section.xpath("./i")[0].text, log_contents[3]
        )

        # Assert lines in the section "A title"
        details = self.assertHasElement(card_body, "./details")
        summary = self.assertHasElement(details, "./summary")
        label = self.assertHasElement(
            summary, './/span[@class="log-section-label"]'
        )
        self.assertTextContentEqual(label, "A title")

        section_pre = self.assertHasElement(
            details, './pre[@class="formatted-content"]'
        )

        section_lines = section_pre.xpath("./i")
        self.assertEqual(len(section_lines), 2)

        self.assertEqual(section_lines[0].text, log_contents[1])
        self.assertEqual(section_lines[1].text, log_contents[2])

    def test_incomplete(self) -> None:
        self.file.complete = False
        self.file.save()
        widget = LogFileWidget(self.file, sections=[])
        context = widget.get_context_data()

        self.assertEqual(
            context,
            {
                "formatted_content_div_id": "formatted-content",
                "file_ui": FileUI.from_file_in_artifact(self.file),
                "file_in_artifact": self.file,
                "widget": widget,
            },
        )

    def test_add_line_numbers(self) -> None:
        self.assertHTMLEqual(
            LogFileWidget._add_line_numbers(
                [
                    "first line",
                    "second <b>line</b>",
                    "third <b>line</b>",
                    "fourth <b>line</b>",
                ],
                start=2,
                end=3,
                suffix="-big",
            ),
            textwrap.dedent(
                """
                <a id="L2-big" href="#L2-big">2</a>
                <i>second &lt;b&gt;line&lt;/b&gt;</i>
                <a id="L3-big" href="#L3-big">3</a>
                <i>third &lt;b&gt;line&lt;/b&gt;</i>"""
            ),
        )


class LinenoHtmlFormatterTests(TestCase):
    """Tests for LinenoHtmlFormatter class."""

    def test_wrap(self) -> None:
        formatter = LinenoHtmlFormatter(suffix="-suffix")

        with mock.patch(
            "debusine.web.views.files.add_line_number"
        ) as add_line_number_mock:
            self.assertEqual(
                list(formatter.wrap([(0, "<pre>")])), [(0, "<pre>")]
            )

        add_line_number_mock.assert_not_called()

        with mock.patch(
            "debusine.web.views.files.add_line_number", return_value="CHANGED"
        ) as add_line_number_mock:
            self.assertEqual(
                list(formatter.wrap([(1, "foo"), (1, "bar\n")])),
                [(1, "CHANGED"), (1, "CHANGED")],
            )

        self.assertEqual(
            add_line_number_mock.mock_calls,
            [
                mock.call(line_number=1, suffix="-suffix", line="foo"),
                mock.call(line_number=2, suffix="-suffix", line="bar"),
            ],
        )


class RenderedSectionTests(TestCase):
    """Tests for RenderedSection."""

    def test_content_raise_not_safe_string(self) -> None:
        with self.assertRaises(ValidationError) as exc:
            LogFileWidget.RenderedSection(
                start_line=1,
                title=None,
                expanded=True,
                content="<b>unsafe</b>",
            )

        self.assertIn("content", str(exc.exception))

    def test_content_validator(self) -> None:
        # Does not raise
        rendered_section = LogFileWidget.RenderedSection(
            start_line=1,
            title=None,
            expanded=True,
            content=escape("<b>safe</b>"),
        )
        self.assertIsInstance(rendered_section.content, SafeString)


def add_line_numbers(
    *, start: int, lines: list[str], suffix: str = ""
) -> SafeString:
    """Format lines."""
    formatted = [
        add_line_number(
            line_number=line_number,
            suffix=suffix,
            line=SafeString(line),
        )
        for line_number, line in enumerate(lines, start=start)
    ]
    return mark_safe("".join(formatted))



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for base signing task classes."""

from pathlib import Path
from typing import override
from unittest import mock

from django.db import connections
from django.test import TestCase, TransactionTestCase

from debusine.artifacts.models import WorkRequestResults
from debusine.signing.tasks import BaseSigningTask
from debusine.tasks import FALLBACK_ARCHITECTURE
from debusine.tasks.models import BaseDynamicTaskData, BaseTaskData, WorkerType
from debusine.tasks.tests.helpers import (
    ExternalTaskHelperMixin,
    SampleBaseExternalTask,
)


class SampleBaseSigningTask(
    SampleBaseExternalTask[BaseTaskData, BaseDynamicTaskData],
    BaseSigningTask[BaseTaskData, BaseDynamicTaskData],
):
    """Sample class to test BaseSigningTask class."""

    @override
    def run(
        self,
        execute_directory: Path,  # noqa: ARG002, U100
    ) -> WorkRequestResults:
        """Succeed if this task is running in a transaction."""
        return (
            WorkRequestResults.SUCCESS
            if connections["default"].in_atomic_block
            else WorkRequestResults.FAILURE
        )


class SampleBaseSigningTask2Data(BaseTaskData):
    """Data representation for SampleBaseSigningTask2."""

    foo: str


class SampleBaseSigningTask2(
    SampleBaseExternalTask[SampleBaseSigningTask2Data, BaseDynamicTaskData],
    BaseSigningTask[SampleBaseSigningTask2Data, BaseDynamicTaskData],
):
    """Test BaseSigningTask class with jsonschema validation."""

    TASK_VERSION = 1

    @override
    def run(self, execute_directory: Path) -> WorkRequestResults:  # noqa: U100
        """Unused abstract method from BaseExternalTask."""
        raise NotImplementedError()


class BaseSigningTaskTests(TestCase):
    """Unit tests for :py:class:`BaseSigningTask`."""

    @override
    def setUp(self) -> None:
        super().setUp()
        self.task = SampleBaseSigningTask({})
        self.task2 = SampleBaseSigningTask2({"foo": "bar"})
        self.worker_metadata = {
            "system:worker_type": WorkerType.SIGNING,
            "system:architectures": [FALLBACK_ARCHITECTURE],
        }

    def test_can_run_on_no_version(self) -> None:
        """Ensure can_run_on returns True if no version is specified."""
        self.assertIsNone(self.task.TASK_VERSION)
        metadata = {**self.worker_metadata, **self.task.analyze_worker()}
        self.assertEqual(self.task.can_run_on(metadata), True)

    def test_can_run_on_with_different_versions(self) -> None:
        """Ensure can_run_on returns False if versions differ."""
        self.assertIsNone(self.task.TASK_VERSION)
        metadata = {**self.worker_metadata, **self.task.analyze_worker()}
        metadata["signing:samplebasesigningtask:version"] = 1
        self.assertEqual(self.task.can_run_on(metadata), False)


class BaseSigningTaskTransactionTests(
    ExternalTaskHelperMixin[SampleBaseSigningTask], TransactionTestCase
):
    """Test transactional behaviour of :py:class:`BaseSigningTask`."""

    def test_runs_in_transaction(self) -> None:
        """Tasks are executed in a transaction."""
        task = self.configure_task()
        with mock.patch.object(
            task, "_upload_work_request_debug_logs", autospec=True
        ):
            self.assertEqual(task.execute(), WorkRequestResults.SUCCESS)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Workflow to update metadata for a single suite."""

import datetime as dt
from functools import cached_property
from typing import override

from django.db import connection

from debusine.artifacts.models import ArtifactCategory, CollectionCategory
from debusine.client.models import LookupChildType
from debusine.db.models import Collection, WorkRequest
from debusine.server.tasks.models import GenerateSuiteIndexesData
from debusine.server.workflows.base import Workflow
from debusine.server.workflows.models import (
    UpdateSuiteData,
    WorkRequestWorkflowData,
)
from debusine.signing.models import SigningMode
from debusine.signing.tasks.models import SignRepositoryIndexData
from debusine.tasks import ensure_collection_category
from debusine.tasks.models import (
    ActionUpdateCollectionWithArtifacts,
    BaseDynamicTaskData,
)


class UpdateSuiteWorkflow(Workflow[UpdateSuiteData]):
    """Update metadata for a single suite."""

    TASK_NAME = "update_suite"

    def _get_current_transaction_timestamp(self) -> dt.datetime:
        with connection.cursor() as cursor:
            cursor.execute("SELECT CURRENT_TIMESTAMP")
            [now] = cursor.fetchone()
        assert isinstance(now, dt.datetime)
        return now

    @cached_property
    def suite_collection(self) -> Collection:
        """The suite whose indexes should be generated."""
        suite = self.work_request.lookup_single(
            self.data.suite_collection, expect_type=LookupChildType.COLLECTION
        ).collection
        ensure_collection_category(
            configuration_key="suite_collection",
            category=suite.category,
            expected=CollectionCategory.SUITE,
        )
        return suite

    @override
    def compute_dynamic_data(self) -> BaseDynamicTaskData:
        """
        Compute dynamic data for this workflow.

        :subject: name of ``suite_collection``
        """
        return BaseDynamicTaskData(subject=self.suite_collection.name)

    @override
    def populate(self) -> None:
        """Create child work requests."""
        # TODO: It would be useful to be able to run this workflow for a
        # given timestamp in order to backfill history.
        now = self._get_current_transaction_timestamp()
        generate_task, _ = self.work_request_ensure_child_server(
            task_name="generatesuiteindexes",
            task_data=GenerateSuiteIndexesData(
                suite_collection=self.data.suite_collection, generate_at=now
            ),
            workflow_data=WorkRequestWorkflowData(
                display_name=(
                    f"Generate indexes for {self.suite_collection.name}"
                ),
                step=f"generate-suite-indexes-{self.suite_collection.name}",
            ),
        )
        generated_release_name = (
            f"generated-release-{self.suite_collection.name}"
        )
        self.provides_artifact(
            generate_task,
            ArtifactCategory.REPOSITORY_INDEX,
            generated_release_name,
            artifact_filters={"data__path": "Release"},
        )
        generated_release = (
            f"internal@collections/name:{generated_release_name}"
        )

        sign_tasks: dict[str, WorkRequest] = {}
        for mode, signed_name in (
            (SigningMode.DETACHED, "Release.gpg"),
            (SigningMode.CLEAR, "InRelease"),
        ):
            sign_tasks[signed_name], _ = self.work_request_ensure_child_signing(
                task_name="signrepositoryindex",
                task_data=SignRepositoryIndexData(
                    suite_collection=self.data.suite_collection,
                    unsigned=generated_release,
                    mode=mode,
                    signed_name=signed_name,
                ),
                workflow_data=WorkRequestWorkflowData(
                    display_name=(
                        f"Create signed {signed_name} for "
                        f"{self.suite_collection.name}"
                    ),
                    step=f"sign-{mode}-{self.suite_collection.name}",
                ),
            )
            self.requires_artifact(sign_tasks[signed_name], generated_release)
            sign_tasks[signed_name].add_event_reaction(
                "on_success",
                ActionUpdateCollectionWithArtifacts(
                    collection=self.data.suite_collection,
                    variables={"path": signed_name},
                    artifact_filters={
                        "category": ArtifactCategory.REPOSITORY_INDEX
                    },
                    created_at=now,
                ),
            )
        # Make sure InRelease is generated last, in order that the "current"
        # view of a suite can arrange to only show a snapshot once an
        # InRelease file exists for it.
        sign_tasks["InRelease"].add_dependency(sign_tasks["Release.gpg"])



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""API for provisioning and managing Hetzner Cloud workers."""

import logging
from functools import lru_cache
from typing import Any, TYPE_CHECKING, override

import hcloud
from django.utils import timezone

# Before hcloud 1.27.0, we must import from the .domain modules
from hcloud.images.domain import Image
from hcloud.locations.domain import Location
from hcloud.networks.domain import Network
from hcloud.server_types.domain import ServerType
from hcloud.servers.domain import Server, ServerCreatePublicNetwork
from hcloud.ssh_keys.domain import SSHKey

from debusine.assets import CloudProvidersType, HetznerProviderAccountData
from debusine.db.models.auth import Token
from debusine.server.worker_pools.base import WorkerPoolInterface
from debusine.server.worker_pools.cloud_init import (
    CloudInitConfig,
    worker_bootstrap_cloud_init,
)
from debusine.server.worker_pools.exceptions import InstanceLaunchException
from debusine.server.worker_pools.models import (
    HetznerCloudWorkerPoolSpecification,
)

if TYPE_CHECKING:
    from debusine.db.models import Worker

log = logging.getLogger(__name__)


class HetznerCloudWorkerPool(
    WorkerPoolInterface, cloud_provider=CloudProvidersType.HETZNER
):
    """Hetzner Cloud Worker Pool."""

    specifications: HetznerCloudWorkerPoolSpecification
    client: hcloud.Client

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Initialize an HetznerCloudWorkerPool."""
        super().__init__(*args, **kwargs)
        provider_account_data = self.worker_pool.provider_account.data_model
        assert isinstance(provider_account_data, HetznerProviderAccountData)
        self.client = self._make_client(provider_account_data)

    @override
    def launch_worker(self, worker: "Worker") -> None:
        assert worker.activation_token

        self._terminate_existing_instances(worker.worker_pool_data)

        instance_id = self._create_instance(
            worker_name=worker.name, activation_token=worker.activation_token
        )

        worker.instance_created_at = timezone.now()
        worker.worker_pool_data = {"instance-id": instance_id}
        worker.save()

    @override
    def terminate_missing(self) -> None:
        instance_ids: dict[int, int] = {
            instance_id: worker_id
            for (
                worker_id,
                instance_id,
            ) in self.worker_pool.worker_set.active().values_list(
                "id", "worker_pool_data__instance-id"
            )
        }
        if not instance_ids:
            return

        for server in self.client.servers.get_all():
            if server.status in (
                "running",
                "initializing",
                "starting",
                "migrating",
                "rebuilding",
            ):
                assert server.id
                instance_ids.pop(server.id, None)

        for instance_id, worker_id in instance_ids.items():
            worker = self.worker_pool.worker_set.get(id=worker_id)
            log.warning(
                "Worker %s (server %s) has gone missing. Terminating.",
                worker.name,
                instance_id,
            )
            self.worker_pool.terminate_worker(worker)

    def _make_client(
        self, provider_account_data: HetznerProviderAccountData
    ) -> hcloud.Client:
        """Make a Hetzner Cloud client for this provider account."""
        return hcloud.Client(token=provider_account_data.credentials.api_token)

    @override
    def _terminate_existing_instances(
        self, worker_pool_data: dict[str, Any] | None
    ) -> None:
        """If worker_pool_data references an instance-id, terminate it."""
        if worker_pool_data is None:
            return
        if existing_instance := worker_pool_data.get("instance-id", None):
            try:
                self.client.servers.delete(Server(id=existing_instance))
            except hcloud.APIException as e:
                if e.code != "not_found":
                    raise

    def _cloud_init_config(
        self, worker_name: str, activation_token: Token
    ) -> CloudInitConfig:
        """Generate cloud-init configuration for a new instance."""
        return worker_bootstrap_cloud_init(
            worker_name,
            activation_token,
            debusine_install_source=self.specifications.debusine_install_source,
            debusine_install_source_custom=self.specifications.debusine_install_source_custom,
            debian_release=self.specifications.debian_release,
            apt_authentication=self.get_apt_authentication(),
        )

    @lru_cache
    def _get_network_by_name(self, name: str) -> Network:
        """Lookup a Hetzner network by name."""
        network = self.client.networks.get_by_name(name)
        if network is None:
            raise InstanceLaunchException(f"Unable to locate network {name!r}")
        return network

    def _create_instance(
        self, worker_name: str, activation_token: Token
    ) -> int:
        """Create a Hetzner Cloud server, and return its ID."""
        response = self.client.servers.create(
            name=worker_name,
            server_type=ServerType(name=self.specifications.server_type),
            image=Image(name=self.specifications.image_name),
            ssh_keys=[
                SSHKey(name=name) for name in self.specifications.ssh_keys
            ],
            location=(
                Location(name=self.specifications.location)
                if self.specifications.location
                else None
            ),
            networks=[
                self._get_network_by_name(name)
                for name in self.specifications.networks
            ],
            user_data=self._cloud_init_config(
                worker_name, activation_token
            ).as_str(),
            labels=self.specifications.labels,
            public_net=ServerCreatePublicNetwork(
                enable_ipv4=self.specifications.enable_ipv4,
                enable_ipv6=self.specifications.enable_ipv6,
            ),
        )

        assert response.server.id
        return response.server.id



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""The collection manager for debian:suite collections."""

import datetime as dt
import re
from collections.abc import Generator
from contextlib import contextmanager
from typing import Any, override

from debian.debian_support import Version
from django.db import IntegrityError
from django.db.models import QuerySet
from django.db.models.fields.json import KT
from django.utils import timezone
from pgpy import PGPKey

from debusine.artifacts.local_artifact import BinaryPackage, SourcePackage
from debusine.artifacts.models import ArtifactCategory, CollectionCategory
from debusine.assets.asset_utils import ExportablePGPKeyring
from debusine.assets.models import AssetCategory, KeyPurpose
from debusine.db.models import (
    Artifact,
    Asset,
    Collection,
    CollectionItem,
    CollectionItemMatchConstraint,
    User,
    WorkRequest,
)
from debusine.db.models.assets import get_public_keys
from debusine.server.collections.base import (
    CollectionManagerInterface,
    ItemAdditionError,
)


def make_source_prefix(source_package_name: str) -> str:
    """
    Make the prefix for a source package name in a Debian repository's pool.

    This is the "<first letter of source name>" / "<first 4 letters of
    source name>" segment, as described in
    https://lists.debian.org/debian-devel/2000/10/msg01340.html.
    """
    if source_package_name.startswith("lib"):
        return source_package_name[:4]
    else:
        return source_package_name[:1]


def make_pool_filename(
    source_package_name: str, component: str, base_name: str
) -> str:
    """
    Make a file name in a Debian repository's pool.

    See https://lists.debian.org/debian-devel/2000/10/msg01340.html for an
    explanation of the structure.
    """
    prefix = make_source_prefix(source_package_name)
    return f"pool/{component}/{prefix}/{source_package_name}/{base_name}"


class DebianSuiteManager(CollectionManagerInterface):
    """Manage collection of category debian:suite."""

    COLLECTION_CATEGORY = CollectionCategory.SUITE
    VALID_ARTIFACT_CATEGORIES = frozenset(
        {
            ArtifactCategory.SOURCE_PACKAGE,
            ArtifactCategory.BINARY_PACKAGE,
            ArtifactCategory.REPOSITORY_INDEX,
        }
    )

    @override
    def __init__(self, collection: Collection) -> None:
        super().__init__(collection)
        self.may_rewind_versions = False

    @property
    def parent_archive(self) -> Collection | None:
        """Get the parent archive, if there is one."""
        try:
            return Collection.objects.get(
                workspace=self.collection.workspace,
                category=CollectionCategory.ARCHIVE,
                name="_",
            )
        except Collection.DoesNotExist:
            return None

    @property
    def may_reuse_versions(self) -> bool:
        """True if versions in this suite may be reused."""
        if self.collection.data.get("may_reuse_versions", False):
            return True
        if (archive := self.parent_archive) is None:
            return False
        if archive.data.get("may_reuse_versions", False):
            return True
        return False

    def get_signing_keys(self) -> QuerySet[Asset]:
        """Return signing keys for this suite."""
        if self.collection.data.get("signing_keys") is not None:
            fingerprints = self.collection.data["signing_keys"]
        else:
            if (archive := self.parent_archive) is not None:
                fingerprints = archive.data.get("signing_keys", [])
            else:
                return Asset.objects.none()
        # TODO: This might not be quite right for snapshots; we always just
        # return the current signing keys.  However, at the moment we only
        # track the creation time of assets, but we don't track when asset
        # usage records are created or removed and we don't retain removed
        # assets; so it's hard to do much better.
        return Asset.objects.filter(
            category=AssetCategory.SIGNING_KEY,
            data__purpose=KeyPurpose.OPENPGP_REPOSITORY,
            data__fingerprint__in=fingerprints,
            usage__workspace=self.collection.workspace,
        )

    def export_signing_keys(self) -> str | None:
        """Export signing keys for this suite as an ASCII-armored keyring."""
        signing_keys = self.get_signing_keys()
        if not signing_keys:
            return None
        # PGPy's upstream maintenance status is currently unclear; see e.g.
        # https://github.com/SecurityInnovation/PGPy/issues/462.  Since it's
        # pure-Python the risk is relatively low, but it's still possible
        # that it might be confused into producing misleading keyrings.
        # Think more carefully before allowing untrusted users to upload
        # their own signing keys.
        return str(
            ExportablePGPKeyring(
                *(
                    PGPKey.from_blob(public_key)[0]
                    for public_key in get_public_keys(signing_keys)
                )
            )
        )

    def make_constraints(
        self, item: CollectionItem, source_package_name: str, component: str
    ) -> Generator[CollectionItemMatchConstraint]:
        """Yield constraints for an item added to this collection."""
        assert item.artifact is not None
        containing_archives = Collection.objects.filter(
            category=CollectionCategory.ARCHIVE,
            child_items__removed_at__isnull=True,
            child_items__collection=self.collection,
        )
        for file_in_artifact in item.artifact.fileinartifact_set.all():
            pool_filename = make_pool_filename(
                source_package_name, component, file_in_artifact.path
            )
            file = file_in_artifact.file
            hash_digest = (
                f"{file.current_hash_algorithm}:{file.hash_digest.hex()}"
            )
            # Each file name in pool/ must only refer to at most one
            # concrete file in the suite at a given time.
            yield CollectionItemMatchConstraint(
                collection=self.collection,
                collection_item_id=item.id,
                constraint_name="pool-file",
                key=pool_filename,
                value=hash_digest,
            )
            # The pool/ directory is shared between suites in the same
            # archive, so we also need archive-level constraints to ensure
            # that different suites in the same archive don't contain the
            # same pool file with different contents.
            for archive in containing_archives:
                yield CollectionItemMatchConstraint(
                    collection=archive,
                    collection_item_id=item.id,
                    constraint_name="pool-file",
                    key=pool_filename,
                    value=hash_digest,
                )

    @contextmanager
    def allow_rewinding_versions(self) -> Generator[None]:
        """Temporarily allow rewinding versions in this suite."""
        original_may_rewind_versions = self.may_rewind_versions
        try:
            self.may_rewind_versions = True
            yield
        finally:
            self.may_rewind_versions = original_may_rewind_versions

    def _check_existing_versions(
        self,
        new_version: Version,
        existing_items: QuerySet[CollectionItem],
        *,
        kind: str,
        package_name: str,
    ) -> None:
        if not self.may_rewind_versions and not self.collection.data.get(
            "may_rewind_versions", False
        ):
            for existing_item in existing_items:
                if new_version < Version(existing_item.data["version"]):
                    raise ItemAdditionError(
                        f"{self.collection}: New {kind} version "
                        f"{package_name} {new_version} < "
                        f"existing version {existing_item.data["version"]}"
                    )

    @override
    def do_add_artifact(
        self,
        artifact: Artifact,
        *,
        user: User,
        workflow: WorkRequest | None = None,
        variables: dict[str, Any] | None = None,
        name: str | None = None,
        replace: bool = False,
        created_at: dt.datetime | None = None,
        replaced_by: CollectionItem | None = None,
    ) -> CollectionItem:
        required_variables: dict[str, tuple[str, ...]] = {
            ArtifactCategory.SOURCE_PACKAGE: ("component", "section"),
            ArtifactCategory.BINARY_PACKAGE: (
                "component",
                "section",
                "priority",
            ),
            ArtifactCategory.REPOSITORY_INDEX: ("path",),
        }
        for required_variable in required_variables[artifact.category]:
            if required_variable not in (variables or {}):
                raise ItemAdditionError(
                    f'Adding {artifact.category} to {CollectionCategory.SUITE} '
                    f'requires "{required_variable}"'
                )
        assert variables is not None

        if artifact.category == ArtifactCategory.SOURCE_PACKAGE:
            source_package_data = SourcePackage.create_data(artifact.data)
            data = {
                "package": source_package_data.name,
                "version": source_package_data.version,
                "component": variables["component"],
                "section": variables["section"],
            }
            name = "{package}_{version}".format(**data)
            source_package_name = source_package_data.name
            self._check_existing_versions(
                Version(data["version"]),
                self.collection.child_items.active().filter(
                    child_type=CollectionItem.Types.ARTIFACT,
                    category=ArtifactCategory.SOURCE_PACKAGE,
                    data__package=data["package"],
                ),
                kind="source",
                package_name=data["package"],
            )
        elif artifact.category == ArtifactCategory.BINARY_PACKAGE:
            binary_package_data = BinaryPackage.create_data(artifact.data)
            data = {
                "srcpkg_name": binary_package_data.srcpkg_name,
                "srcpkg_version": binary_package_data.srcpkg_version,
                "package": binary_package_data.deb_fields["Package"],
                "version": binary_package_data.deb_fields["Version"],
                "architecture": binary_package_data.deb_fields["Architecture"],
                "component": variables["component"],
                "section": variables["section"],
                "priority": variables["priority"],
            }
            name = "{package}_{version}_{architecture}".format(**data)
            source_package_name = binary_package_data.srcpkg_name
            self._check_existing_versions(
                Version(data["version"]),
                self.collection.child_items.active().filter(
                    child_type=CollectionItem.Types.ARTIFACT,
                    category=ArtifactCategory.BINARY_PACKAGE,
                    data__package=data["package"],
                    data__architecture=data["architecture"],
                ),
                kind="binary",
                package_name=f"{data['package']}:{data['architecture']}",
            )
        else:
            assert artifact.category == ArtifactCategory.REPOSITORY_INDEX
            data = {"path": variables["path"]}
            name = "index:{path}".format(**data)

        if replace:
            self.remove_items_by_name(
                name=name,
                child_types=[CollectionItem.Types.ARTIFACT],
                user=user,
                workflow=workflow,
            )

        try:
            item = CollectionItem.objects.get(
                parent_collection=self.collection,
                name=name,
                artifact=artifact,
                child_type=CollectionItem.Types.ARTIFACT,
                category=artifact.category,
                data=data,
            )
        except CollectionItem.DoesNotExist:
            pass
        else:
            # The same artifact is already in the collection with the same
            # name and data.
            return item

        try:
            item = CollectionItem.objects.create_from_artifact(
                artifact,
                parent_collection=self.collection,
                name=name,
                data=data,
                created_at=created_at,
                created_by_user=user,
                created_by_workflow=workflow,
                replaced_by=replaced_by,
            )
            if replaced_by is None and artifact.category in {
                ArtifactCategory.SOURCE_PACKAGE,
                ArtifactCategory.BINARY_PACKAGE,
            }:
                CollectionItemMatchConstraint.objects.bulk_create(
                    self.make_constraints(
                        item, source_package_name, data["component"]
                    )
                )
            return item
        except IntegrityError as exc:
            raise ItemAdditionError(str(exc))

    @override
    def do_remove_item(
        self,
        item: CollectionItem,
        *,
        user: User | None = None,
        workflow: WorkRequest | None = None,
    ) -> None:
        item.removed_by_user = user
        item.removed_by_workflow = workflow
        item.removed_at = timezone.now()
        item.save()
        if (
            item.artifact is not None
            and (
                item.artifact.category
                in {
                    ArtifactCategory.SOURCE_PACKAGE,
                    ArtifactCategory.BINARY_PACKAGE,
                }
            )
            and self.may_reuse_versions
        ):
            # This deletes both suite and archive constraints.
            CollectionItemMatchConstraint.objects.filter(
                collection_item_id=item.id, constraint_name="pool-file"
            ).delete()

    @override
    def do_lookup(self, query: str) -> CollectionItem | None:
        """
        Return one CollectionItem based on the query.

        :param query: `source:NAME`, `source-version:NAME_VERSION`,
          `binary:NAME_ARCHITECTURE`,
          `binary-version:NAME_VERSION_ARCHITECTURE`, or `index:PATH`.  If
          more than one possible CollectionItem matches the query (which is
          possible for `source:` and `binary:` queries): return the one with
          the highest version.
        """
        qs = self.collection.child_items.active().filter(
            child_type=CollectionItem.Types.ARTIFACT,
            # Technically redundant, but helps PostgreSQL to use the correct
            # index where available (e.g. db_ci_suite_source_idx).
            parent_category=CollectionCategory.SUITE,
        )
        sort_by_version = True

        if m := re.match(r"^source:(.+)$", query):
            qs = qs.annotate(package=KT("data__package")).filter(
                category=ArtifactCategory.SOURCE_PACKAGE,
                package=m.group(1),
            )
        elif m := re.match(r"^source-version:(.+?)_(.+)$", query):
            qs = qs.annotate(
                package=KT("data__package"), version=KT("data__version")
            ).filter(
                category=ArtifactCategory.SOURCE_PACKAGE,
                package=m.group(1),
                version=m.group(2),
            )
        elif m := re.match(r"^binary:(.+?)_(.+)$", query):
            qs = qs.annotate(
                package=KT("data__package"),
                architecture=KT("data__architecture"),
            ).filter(
                category=ArtifactCategory.BINARY_PACKAGE,
                package=m.group(1),
                architecture__in={m.group(2), "all"},
            )
        elif m := re.match(r"^binary-version:(.+?)_(.+?)_(.+)$", query):
            qs = qs.annotate(
                package=KT("data__package"),
                version=KT("data__version"),
                architecture=KT("data__architecture"),
            ).filter(
                category=ArtifactCategory.BINARY_PACKAGE,
                package=m.group(1),
                version=m.group(2),
                architecture__in={m.group(3), "all"},
            )
        elif m := re.match(r"^index:(.+)$", query):
            qs = qs.annotate(path=KT("data__path")).filter(
                category=ArtifactCategory.REPOSITORY_INDEX,
                path=m.group(1),
            )
            sort_by_version = False
        else:
            raise LookupError(f'Unexpected lookup format: "{query}"')

        # Ideally the sorting would be done in the database, but that
        # requires the debversion extension and installing that requires a
        # superuser, which is cumbersome since it means we can't do it in a
        # Django migration as currently configured.  We don't expect a
        # single suite to contain many active versions of a single package,
        # so doing the sorting in Python should be fine in practice.
        items = list(qs)
        if sort_by_version:
            items = sorted(
                items, key=lambda item: Version(item.data["version"])
            )
        if not items:
            return None
        return items[-1]



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Test the Dummy Worker Pool Implementation."""

from debusine.server.worker_pools import DummyWorkerPool
from debusine.test.django import TestCase


class DummyWorkerPoolTests(TestCase):
    """Tests for DummyWorkerPool."""

    def test_launch_worker(self) -> None:
        worker_pool = self.playground.create_worker_pool()
        worker_pool_api = DummyWorkerPool(worker_pool=worker_pool)

        worker = self.playground.create_worker(worker_pool=worker_pool)
        worker.instance_created_at = None
        worker.save()

        worker_pool_api.launch_worker(worker)

        self.assertIsNotNone(worker.instance_created_at)
        self.assertEqual(worker.worker_pool_data, {"launched": True})

    def test_terminate_worker(self) -> None:
        worker_pool = self.playground.create_worker_pool()
        worker_pool_api = DummyWorkerPool(worker_pool=worker_pool)

        worker = self.playground.create_worker(worker_pool=worker_pool)
        worker.worker_pool_data = {"something": "Foo"}
        worker.save()

        worker_pool_api.terminate_worker(worker)

        self.assertIsNone(worker.instance_created_at)
        self.assertIsNone(worker.worker_pool_data)

    def test_terminate_missing(self) -> None:
        worker_pool = self.playground.create_worker_pool()
        worker_pool_api = DummyWorkerPool(worker_pool=worker_pool)
        worker_pool_api.terminate_missing()



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for the collections views."""

import datetime as dt
from typing import Any, ClassVar, assert_never, override
from unittest import mock

from django.db import IntegrityError
from django.db.models import Max
from django.urls import reverse
from rest_framework import status

from debusine.artifacts.models import (
    BareDataCategory,
    CollectionCategory,
    DebusineTaskConfiguration,
    TaskTypes,
)
from debusine.client.exceptions import DebusineError
from debusine.client.models import CollectionData, CollectionDataNew
from debusine.db.models import (
    Collection,
    CollectionItem,
    CollectionTokenGrant,
    Token,
    Workspace,
)
from debusine.db.playground import scenarios
from debusine.server.collections import DebusineTaskConfigurationManager
from debusine.server.views.collections import TaskConfigurationCollectionView
from debusine.server.views.tests.base import TestCase
from debusine.test.django import (
    AllowAll,
    DenyAll,
    TestResponseType,
    override_permission,
)


class TaskConfigurationCollectionViewTests(TestCase):
    """Tests for TaskConfigurationCollectionView."""

    scenario = scenarios.DefaultContextAPI()
    collection: ClassVar[Collection]
    manager: ClassVar[DebusineTaskConfigurationManager]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.collection = cls.scenario.default_task_configuration_collection
        cls.manager = DebusineTaskConfigurationManager(
            collection=cls.collection
        )

    @classmethod
    def add_config(cls, entry: DebusineTaskConfiguration) -> CollectionItem:
        """Add a config entry to config_collection."""
        return cls.manager.add_bare_data(
            BareDataCategory.TASK_CONFIGURATION,
            user=cls.scenario.user,
            data=entry,
        )

    def get_items(self) -> list[DebusineTaskConfiguration]:
        collection = self.collection

        return [
            DebusineTaskConfiguration(**child.data)
            for child in collection.child_items.active().filter(
                child_type=CollectionItem.Types.BARE
            )
        ]

    def get(
        self,
        name: str = "default",
        workspace: Workspace | str | None = None,
        user_token: bool = True,
    ) -> TestResponseType:
        """GET request for a named collection."""
        match workspace:
            case None:
                workspace_name = self.scenario.workspace.name
            case str():
                workspace_name = workspace
            case _:
                workspace_name = workspace.name

        match user_token:
            case True:
                headers = {"token": self.scenario.user_token.key}
            case False:
                headers = {}
            case _ as unreachable:
                assert_never(unreachable)

        return self.client.get(
            reverse(
                "api:task-configuration-collection",
                kwargs={"workspace": workspace_name, "name": name},
            ),
            content_type="application/json",
            headers=headers,
        )

    def post(
        self,
        name: str = "default",
        workspace: Workspace | str | None = None,
        user_token: bool = True,
        data: dict[str, Any] | None = None,
    ) -> TestResponseType:
        """POST request for a named collection."""
        match workspace:
            case None:
                workspace_name = self.scenario.workspace.name
            case str():
                workspace_name = workspace
            case _:
                workspace_name = workspace.name

        match user_token:
            case True:
                headers = {"token": self.scenario.user_token.key}
            case False:
                headers = {}
            case _ as unreachable:
                assert_never(unreachable)

        if data is None:
            data = {}

        return self.client.post(
            reverse(
                "api:task-configuration-collection",
                kwargs={"workspace": workspace_name, "name": name},
            ),
            data=data,
            content_type="application/json",
            headers=headers,
        )

    def test_unauthenticated(self) -> None:
        """Authentication is required."""
        for method in "get", "post":
            with self.subTest(method=method):
                response = getattr(self, method)(user_token=False)
                self.assertResponseProblem(
                    response,
                    "Error",
                    detail_pattern=(
                        "Authentication credentials were not provided."
                    ),
                    status_code=status.HTTP_403_FORBIDDEN,
                )

    def test_workspace_not_found(self) -> None:
        """Workspace must exist."""
        for method in "get", "post":
            with self.subTest(method=method):
                response = getattr(self, method)(workspace="does-not-exist")
                self.assertResponseProblem(
                    response,
                    "Workspace not found",
                    detail_pattern=(
                        "Workspace does-not-exist not found in scope debusine"
                    ),
                    status_code=status.HTTP_404_NOT_FOUND,
                )

    def test_workspace_not_accessible(self) -> None:
        """User must be able to display the workspace."""
        workspace = self.playground.create_workspace(name="private")
        for method in "get", "post":
            with self.subTest(method=method):
                response = getattr(self, method)(workspace=workspace)
                self.assertResponseProblem(
                    response,
                    "Workspace not found",
                    detail_pattern=(
                        "Workspace private not found in scope debusine"
                    ),
                    status_code=status.HTTP_404_NOT_FOUND,
                )

    def test_not_found(self) -> None:
        """The collection must exist."""
        for method in "get", "post":
            with self.subTest(method=method):
                response = getattr(self, method)(name="does-not-exist")
                self.assertResponseProblem(
                    response,
                    "Collection not found",
                    detail_pattern=(
                        "Task configuration collection 'does-not-exist'"
                        " does not exist in workspace 'System'"
                    ),
                    status_code=status.HTTP_404_NOT_FOUND,
                )

    def test_get_unauthorized(self) -> None:
        """The user must be able to display the collection."""
        with override_permission(Collection, "can_display", DenyAll):
            response = self.get()
            self.assertResponseProblem(
                response,
                (
                    "playground cannot display collection"
                    " default@debusine:task-configuration"
                ),
                status_code=status.HTTP_403_FORBIDDEN,
            )

    def test_post_unauthorized(self) -> None:
        """The user must be able to display the collection."""
        with override_permission(
            Workspace, "can_edit_task_configuration", DenyAll
        ):
            response = self.post()
            self.assertResponseProblem(
                response,
                (
                    "playground cannot edit task configuration"
                    " in debusine/System"
                ),
                status_code=status.HTTP_403_FORBIDDEN,
            )

    def test_get(self) -> None:
        response = self.get()
        data = self.assertAPIResponseOk(response)

        collection = data["collection"]
        self.assertEqual(collection["id"], self.collection.pk)
        self.assertEqual(collection["name"], "default")
        self.assertEqual(collection["data"], self.collection.data)

        items = data["items"]
        self.assertEqual(items, [])

    def test_get_items(self) -> None:
        item = DebusineTaskConfiguration(
            task_type=TaskTypes.WORKER, task_name="noop"
        )
        self.add_config(item)
        response = self.get()
        data = self.assertAPIResponseOk(response)
        items = data["items"]
        self.assertEqual(items, [item.model_dump()])

    def test_post_matches_collection(self) -> None:
        self.playground.create_group_role(
            self.scenario.workspace, Workspace.Roles.OWNER, [self.scenario.user]
        )
        for data, detail_pattern in (
            (
                {"id": 0, "name": self.collection.name, "data": {}},
                r"Data posted for collection 0"
                rf" to update collection {self.collection.pk}",
            ),
            (
                {
                    "id": self.collection.pk,
                    "name": "does-not-exist",
                    "data": {},
                },
                r"Data posted for collection 'does-not-exist'"
                r" to update collection 'default'",
            ),
        ):
            with self.subTest(data=data):
                response = self.post(data={"collection": data, "items": []})
                self.assertResponseProblem(
                    response,
                    "Posted collection mismatch",
                    detail_pattern=detail_pattern,
                    status_code=status.HTTP_400_BAD_REQUEST,
                )

    def test_post_add(self) -> None:
        self.playground.create_group_role(
            self.scenario.workspace, Workspace.Roles.OWNER, [self.scenario.user]
        )

        item = DebusineTaskConfiguration(
            task_type=TaskTypes.WORKER, task_name="noop"
        )
        response = self.post(
            data={
                "collection": {
                    "id": self.collection.pk,
                    "name": self.collection.name,
                    "data": self.collection.data,
                },
                "items": [item.model_dump()],
            }
        )
        data = self.assertAPIResponseOk(response)
        self.assertEqual(
            data, {"added": 1, "removed": 0, "updated": 0, "unchanged": 0}
        )

        self.assertEqual(self.get_items(), [item])

    def test_post_remove(self) -> None:
        self.add_config(
            DebusineTaskConfiguration(
                task_type=TaskTypes.WORKER, task_name="noop"
            )
        )
        self.playground.create_group_role(
            self.scenario.workspace, Workspace.Roles.OWNER, [self.scenario.user]
        )

        response = self.post(
            data={
                "collection": {
                    "id": self.collection.pk,
                    "name": self.collection.name,
                    "data": self.collection.data,
                },
                "items": [],
            }
        )
        data = self.assertAPIResponseOk(response)
        self.assertEqual(
            data, {"added": 0, "removed": 1, "updated": 0, "unchanged": 0}
        )

        self.assertEqual(self.get_items(), [])

    def test_post_update(self) -> None:
        old = DebusineTaskConfiguration(
            task_type=TaskTypes.WORKER, task_name="noop"
        )
        new = DebusineTaskConfiguration(
            task_type=TaskTypes.WORKER,
            task_name="noop",
            comment="updated",
        )
        self.add_config(old)
        self.playground.create_group_role(
            self.scenario.workspace, Workspace.Roles.OWNER, [self.scenario.user]
        )

        response = self.post(
            data={
                "collection": {
                    "id": self.collection.pk,
                    "name": self.collection.name,
                    "data": self.collection.data,
                },
                "items": [new.model_dump()],
            }
        )
        data = self.assertAPIResponseOk(response)
        self.assertEqual(
            data, {"added": 0, "removed": 0, "updated": 1, "unchanged": 0}
        )
        self.assertEqual(self.get_items(), [new])

    def test_post_dry_run(self) -> None:
        old = DebusineTaskConfiguration(
            task_type=TaskTypes.WORKER, task_name="noop"
        )
        new = DebusineTaskConfiguration(
            task_type=TaskTypes.WORKER,
            task_name="noop",
            comment="updated",
        )
        self.add_config(old)
        self.playground.create_group_role(
            self.scenario.workspace, Workspace.Roles.OWNER, [self.scenario.user]
        )

        response = self.post(
            data={
                "collection": {
                    "id": self.collection.pk,
                    "name": self.collection.name,
                    "data": self.collection.data,
                },
                "items": [new.model_dump()],
                "dry_run": True,
            }
        )
        data = self.assertAPIResponseOk(response)
        self.assertEqual(
            data, {"added": 0, "removed": 0, "updated": 1, "unchanged": 0}
        )
        self.assertEqual(self.get_items(), [old])

    def test_post_git_commit_invalid_type(self) -> None:
        self.playground.create_group_role(
            self.scenario.workspace, Workspace.Roles.OWNER, [self.scenario.user]
        )
        value: Any
        for value in (False, 42, [], {}):
            with self.subTest(value=value):
                response = self.post(
                    data={
                        "collection": {
                            "id": self.collection.pk,
                            "name": self.collection.name,
                            "data": {"git_commit": value},
                        },
                        "items": [],
                        "dry_run": True,
                    }
                )
                self.assertResponseProblem(
                    response,
                    "Invalid git commit",
                    detail_pattern="git_commit is not a string",
                    status_code=status.HTTP_400_BAD_REQUEST,
                )

    def test_post_git_commit_invalid_string(self) -> None:
        self.playground.create_group_role(
            self.scenario.workspace, Workspace.Roles.OWNER, [self.scenario.user]
        )
        for value in ("f00", "foo", "", "a" * 100):
            with self.subTest(value=value):
                response = self.post(
                    data={
                        "collection": {
                            "id": self.collection.pk,
                            "name": self.collection.name,
                            "data": {"git_commit": value},
                        },
                        "items": [],
                        "dry_run": True,
                    }
                )
                self.assertResponseProblem(
                    response,
                    "Invalid git commit",
                    detail_pattern="git_commit is malformed",
                    status_code=status.HTTP_400_BAD_REQUEST,
                )

    def test_post_valid_git_commit(self) -> None:
        self.playground.create_group_role(
            self.scenario.workspace, Workspace.Roles.OWNER, [self.scenario.user]
        )
        commit = "bb07924dd93295740939e66db5a23777439c7c51"
        self.collection.data["git_commit"] = "0" * 40
        self.collection.save()

        data = self.collection.data.copy()
        data["git_commit"] = commit
        response = self.post(
            data={
                "collection": {
                    "id": self.collection.pk,
                    "name": self.collection.name,
                    "data": data,
                },
                "items": [],
                "dry_run": True,
            }
        )
        data = self.assertAPIResponseOk(response)
        self.assertEqual(
            data, {"added": 0, "removed": 0, "updated": 0, "unchanged": 0}
        )

        self.collection.refresh_from_db()
        self.assertEqual(self.collection.data["git_commit"], commit)

    def test_diff_items(self) -> None:
        item1 = DebusineTaskConfiguration(
            task_type=TaskTypes.WORKER, task_name="noop", subject="item1"
        )
        item2 = DebusineTaskConfiguration(
            task_type=TaskTypes.WORKER, task_name="noop", subject="item2"
        )
        item3 = DebusineTaskConfiguration(
            task_type=TaskTypes.WORKER, task_name="noop", subject="item3"
        )
        item4 = DebusineTaskConfiguration(
            task_type=TaskTypes.WORKER, task_name="noop", subject="item4"
        )

        for old_items, new_items, removed, added, unchanged in (
            ([], [], [], [], 0),
            ([item1], [], [item1], [], 0),
            ([item1], [item2], [item1], [item2], 0),
            (
                [item1, item2, item3, item4],
                [item2, item4],
                [item1, item3],
                [],
                2,
            ),
            ([item1, item3], [item2, item4], [item1, item3], [item2, item4], 0),
            ([item1, item3], [item3, item4], [item1], [item4], 1),
        ):
            with self.subTest(
                old=[x.subject for x in old_items],
                new=[x.subject for x in new_items],
            ):
                # Populate the collection with old_items
                self.collection.child_items.all().delete()
                for item in old_items:
                    self.add_config(item)

                actual_removed, actual_added, actual_unchanged = (
                    TaskConfigurationCollectionView._diff_items(
                        self.collection, new_items
                    )
                )
                self.assertEqual(
                    (
                        sorted(x.name for x in actual_removed),
                        sorted(x.name() for x in actual_added),
                        actual_unchanged,
                    ),
                    (
                        sorted(x.name() for x in removed),
                        sorted(x.name() for x in added),
                        unchanged,
                    ),
                )


class CollectionViewSetTests(TestCase):
    """Tests for :py:class:`CollectionViewSet`."""

    scenario = scenarios.DefaultContextAPI()
    sid: ClassVar[Collection]
    sid_links: ClassVar[dict[str, str | dict[str, str]]]
    trixie: ClassVar[Collection]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.sid = cls.playground.create_collection(
            "sid", CollectionCategory.SUITE, data={"test": 1}
        )
        cls.trixie = cls.playground.create_collection(
            "trixie", CollectionCategory.SUITE, data={"test": 1}
        )
        cls.sid_links = cls.get_links(cls.sid)

    @classmethod
    def get_links(
        cls, collection: Collection
    ) -> dict[str, str | dict[str, str]]:
        return {
            "self": reverse(
                "api:collection-detail",
                kwargs={
                    "workspace": collection.workspace.name,
                    "pk": collection.pk,
                },
            ),
            "webui_self": collection.get_absolute_url(),
            "webui_category": {
                "href": reverse(
                    "workspaces:collections:category_list",
                    kwargs={
                        "wname": collection.workspace.name,
                        "ccat": collection.category,
                    },
                ),
                "type": "category",
            },
        }

    def _workspace_url_kwarg(self, workspace: Workspace | str | None) -> str:
        match workspace:
            case None:
                return self.scenario.workspace.name
            case str():
                return workspace
            case _:
                return workspace.name

    def _reverse_list(
        self,
        workspace: Workspace | str | None,
        **kwargs: str,
    ) -> str:
        """Build the URL for a test request."""
        kwargs["workspace"] = self._workspace_url_kwarg(workspace)
        return reverse("api:collection-list", kwargs=kwargs)

    def _reverse_lookup(
        self,
        workspace: Workspace | str | None,
        **kwargs: str,
    ) -> str:
        """Build the URL for a test request."""
        kwargs["workspace"] = self._workspace_url_kwarg(workspace)
        return reverse("api:collection-lookup", kwargs=kwargs)

    def _reverse_detail(
        self,
        collection: int | Collection | None,
        workspace: Workspace | str | None,
        **kwargs: str,
    ) -> str:
        """Build the URL for a test request."""
        kwargs["workspace"] = self._workspace_url_kwarg(workspace)

        match collection:
            case None:
                kwargs["pk"] = str(self.sid.pk)
            case int():
                kwargs["pk"] = str(collection)
            case Collection():
                kwargs["pk"] = str(collection.pk)
            case _ as unreachable:
                assert_never(unreachable)

        return reverse("api:collection-detail", kwargs=kwargs)

    def _headers(self, user_token: bool) -> dict[str, str]:
        """Build headers for a test request."""
        match user_token:
            case True:
                return {"token": self.scenario.user_token.key}
            case False:
                return {}
            case _ as unreachable:
                assert_never(unreachable)

    def lookup(
        self,
        category: str = CollectionCategory.SUITE,
        name: str = "sid",
        workspace: Workspace | str | None = None,
        user_token: bool = True,
    ) -> TestResponseType:
        """GET request for the lookup endpoint."""
        return self.client.get(
            self._reverse_lookup(workspace, category=category, name=name),
            content_type="application/json",
            headers=self._headers(user_token),
        )

    def list(
        self,
        workspace: Workspace | str | None = None,
        user_token: bool = True,
    ) -> TestResponseType:
        """GET request for a named collection."""
        return self.client.get(
            self._reverse_list(workspace),
            content_type="application/json",
            headers=self._headers(user_token),
        )

    def get(
        self,
        collection: int | Collection | None = None,
        workspace: Workspace | str | None = None,
        user_token: bool = True,
    ) -> TestResponseType:
        """GET request for a named collection."""
        return self.client.get(
            self._reverse_detail(collection, workspace),
            content_type="application/json",
            headers=self._headers(user_token),
        )

    def post(
        self,
        workspace: Workspace | str | None = None,
        user_token: bool = True,
        data: dict[str, Any] | None = None,
    ) -> TestResponseType:
        """POST request."""
        return self.client.post(
            self._reverse_list(workspace),
            data=data or {},
            content_type="application/json",
            headers=self._headers(user_token),
        )

    def put(
        self,
        collection: int | Collection | None = None,
        workspace: Workspace | str | None = None,
        user_token: bool = True,
        data: dict[str, Any] | None = None,
    ) -> TestResponseType:
        """PUT request for a named collection."""
        return self.client.put(
            self._reverse_detail(collection, workspace),
            data=data or {},
            content_type="application/json",
            headers=self._headers(user_token),
        )

    def patch(
        self,
        collection: int | Collection | None = None,
        workspace: Workspace | str | None = None,
        user_token: bool = True,
        data: dict[str, Any] | None = None,
    ) -> TestResponseType:
        """PATCH request for a named collection."""
        return self.client.patch(
            self._reverse_detail(collection, workspace),
            data=data or {},
            content_type="application/json",
            headers=self._headers(user_token),
        )

    def delete(
        self,
        collection: int | Collection | None = None,
        workspace: Workspace | str | None = None,
        user_token: bool = True,
    ) -> TestResponseType:
        """DELETE request for a named collection."""
        return self.client.delete(
            self._reverse_detail(collection, workspace),
            headers=self._headers(user_token),
        )

    def assertSidUnchanged(self) -> None:
        """Ensure self.sid is unchanged."""
        self.sid.refresh_from_db()
        self.assertEqual(self.sid.name, "sid")
        self.assertEqual(self.sid.category, CollectionCategory.SUITE)
        self.assertIsNone(self.sid.full_history_retention_period)
        self.assertIsNone(self.sid.metadata_only_retention_period)
        self.assertEqual(self.sid.data, {"test": 1})

    def test_unauthenticated(self) -> None:
        """Authentication is required."""
        for method in ("lookup", "list", "post", "put", "patch", "delete"):
            with self.subTest(method=method):
                response = getattr(self, method)(user_token=False)
                self.assertResponseProblem(
                    response,
                    "Error",
                    detail_pattern=(
                        "Authentication credentials were not provided."
                    ),
                    status_code=status.HTTP_403_FORBIDDEN,
                )

    def test_workspace_not_found(self) -> None:
        """Workspace must exist."""
        for method in ("lookup", "list", "post", "put", "patch", "delete"):
            with self.subTest(method=method):
                response = getattr(self, method)(workspace="does-not-exist")
                self.assertResponseProblem(
                    response,
                    "Workspace not found",
                    detail_pattern=(
                        "Workspace does-not-exist not found in scope debusine"
                    ),
                    status_code=status.HTTP_404_NOT_FOUND,
                )

    def test_workspace_not_accessible(self) -> None:
        """User must be able to display the workspace."""
        workspace = self.playground.create_workspace(name="private")
        for method in ("lookup", "list", "post", "put", "patch", "delete"):
            with self.subTest(method=method):
                response = getattr(self, method)(workspace=workspace)
                self.assertResponseProblem(
                    response,
                    "Workspace not found",
                    detail_pattern=(
                        "Workspace private not found in scope debusine"
                    ),
                    status_code=status.HTTP_404_NOT_FOUND,
                )

    def test_not_found(self) -> None:
        """The collection must exist."""
        max_id = Collection.objects.aggregate(Max('id'))['id__max']
        for method in ("get", "put", "patch", "delete"):
            with self.subTest(method=method):
                response = getattr(self, method)(collection=max_id + 1)
                self.assertResponseProblem(
                    response,
                    "Object not found",
                    status_code=status.HTTP_404_NOT_FOUND,
                )

    def test_detail_unauthorized(self) -> None:
        """The user must be able to display the collection."""
        for method in ("get", "put", "patch", "delete"):
            with (
                self.subTest(method=method),
                override_permission(Collection, "can_display", DenyAll),
            ):
                response = getattr(self, method)()
                self.assertResponseProblem(
                    response,
                    "Object not found",
                    status_code=status.HTTP_404_NOT_FOUND,
                )

    def test_create_unauthorized(self) -> None:
        """The user must be able to create the collection."""
        with override_permission(Workspace, "can_create_collections", DenyAll):
            response = getattr(self, "post")()
            self.assertResponseProblem(
                response,
                "playground cannot create collections"
                f" in {self.scenario.workspace}",
                status_code=status.HTTP_403_FORBIDDEN,
            )

    def test_update_unauthorized(self) -> None:
        """The user must be able to update the collection."""
        for method in ("put", "patch"):
            with (
                self.subTest(method=method),
                override_permission(Collection, "can_configure", DenyAll),
            ):
                response = getattr(self, method)()
                self.assertResponseProblem(
                    response,
                    f"playground cannot configure collection {self.sid}",
                    status_code=status.HTTP_403_FORBIDDEN,
                )

    def test_delete_unauthorized(self) -> None:
        """The user must be able to delete the collection."""
        with override_permission(Collection, "can_delete", DenyAll):
            response = self.delete()
            self.assertResponseProblem(
                response,
                f"playground cannot delete collection {self.sid}",
                status_code=status.HTTP_403_FORBIDDEN,
            )

    def test_lookup(self) -> None:
        response = self.lookup()
        data = self.assertAPIResponseOk(response)
        assert data == {
            "id": self.sid.pk,
            "name": self.sid.name,
            "category": self.sid.category,
            "full_history_retention_period": None,
            "metadata_only_retention_period": None,
            "data": {"test": 1},
            "links": self.sid_links,
            "stats": None,
        }

    def test_lookup_not_found(self) -> None:
        response = self.lookup(category=CollectionCategory.SUITE, name="buster")
        self.assertResponseProblem(
            response,
            "Object not found",
            status_code=status.HTTP_404_NOT_FOUND,
        )

    def test_list(self) -> None:
        response = self.list()
        data = self.assertAPIResponseOk(response)
        names = [f"{x["name"]}@{x["category"]}" for x in data["results"]]
        assert names == [
            '_@debian:archive',
            "_@debian:package-build-logs",
            "sid@debian:suite",
            'trixie@debian:suite',
            'default@debusine:task-configuration',
            '_@debusine:task-history',
        ]

    def test_list_omits_workflow_internal(self) -> None:
        self.playground.create_collection(
            name="1234", category=CollectionCategory.WORKFLOW_INTERNAL
        )
        response = self.list()
        data = self.assertAPIResponseOk(response)
        names = [f"{x["name"]}@{x["category"]}" for x in data["results"]]
        assert names == [
            '_@debian:archive',
            "_@debian:package-build-logs",
            "sid@debian:suite",
            'trixie@debian:suite',
            'default@debusine:task-configuration',
            '_@debusine:task-history',
        ]

    def test_detail(self) -> None:
        response = self.get()
        data = self.assertAPIResponseOk(response)
        assert data == {
            "id": self.sid.pk,
            "name": self.sid.name,
            "category": self.sid.category,
            "full_history_retention_period": None,
            "metadata_only_retention_period": None,
            "data": {"test": 1},
            "links": self.sid_links,
            "stats": None,
        }

    @override_permission(Workspace, "can_create_collections", AllowAll)
    def test_post(self) -> None:
        response = self.post(
            data={
                "category": CollectionCategory.SUITE,
                "name": "new",
                "full_history_retention_period": None,
                "metadata_only_retention_period": None,
                "data": {"created": True},
            }
        )
        new_data = self.assertAPIResponseOk(
            response, status_code=status.HTTP_201_CREATED
        )
        new = Collection.objects.get(
            workspace=self.scenario.workspace,
            category=CollectionCategory.SUITE,
            name="new",
        )
        self.assertIsNone(new.full_history_retention_period)
        self.assertIsNone(new.metadata_only_retention_period)
        self.assertEqual(new.data, {"created": True})
        assert new_data == {
            "id": new.pk,
            "category": CollectionCategory.SUITE,
            "name": "new",
            "full_history_retention_period": None,
            "metadata_only_retention_period": None,
            "data": {"created": True},
            "links": self.get_links(new),
            "stats": None,
        }

    @override_permission(Workspace, "can_create_collections", AllowAll)
    def test_post_name_conflict(self) -> None:
        response = self.post(
            data={
                "category": CollectionCategory.SUITE,
                "name": "sid",
                "full_history_retention_period": None,
                "metadata_only_retention_period": None,
                "data": {},
            }
        )
        self.assertResponseProblem(
            response,
            "Cannot create collection",
            status_code=status.HTTP_400_BAD_REQUEST,
            detail_pattern="A collection with the same name already exists",
        )
        self.assertSidUnchanged()

    @override_permission(Workspace, "can_create_collections", AllowAll)
    def test_post_unexpected_error(self) -> None:
        with mock.patch(
            "debusine.server.serializers.CollectionSerializer.save",
            side_effect=IntegrityError("unexpected message"),
        ):
            response = self.post(
                data={
                    "category": CollectionCategory.SUITE,
                    "name": "sid",
                    "full_history_retention_period": None,
                    "metadata_only_retention_period": None,
                    "data": {},
                }
            )
        self.assertResponseProblem(
            response,
            "Cannot create collection",
            status_code=status.HTTP_400_BAD_REQUEST,
        )
        self.assertSidUnchanged()

    @override_permission(Workspace, "can_create_collections", AllowAll)
    def test_post_invalid_name(self) -> None:
        response = self.post(
            data={
                "category": CollectionCategory.SUITE,
                "name": "sid/test",
                "full_history_retention_period": None,
                "metadata_only_retention_period": None,
                "data": {},
            }
        )
        self.assertResponseProblem(
            response,
            "Cannot deserialize collection",
            validation_errors_pattern=(
                "'sid/test' is not a valid collection name"
            ),
            status_code=status.HTTP_400_BAD_REQUEST,
        )
        self.assertFalse(Collection.objects.filter(name="sid/test").exists())

    @override_permission(Collection, "can_configure", AllowAll)
    def test_patch_noop(self) -> None:
        response = self.patch(data={})
        new_data = self.assertAPIResponseOk(response)
        assert new_data == {
            "id": self.sid.pk,
            "category": self.sid.category,
            "name": self.sid.name,
            "full_history_retention_period": None,
            "metadata_only_retention_period": None,
            "data": self.sid.data,
            "links": self.sid_links,
            "stats": None,
        }
        self.assertSidUnchanged()

    @override_permission(Collection, "can_configure", AllowAll)
    def test_patch_noop_all_set(self) -> None:
        response = self.patch(
            data={
                "name": "sid",
                "full_history_retention_period": None,
                "metadata_only_retention_period": None,
                "data": {"test": 1},
            },
        )
        new_data = self.assertAPIResponseOk(response)
        assert new_data == {
            "id": self.sid.pk,
            "category": self.sid.category,
            "name": self.sid.name,
            "full_history_retention_period": None,
            "metadata_only_retention_period": None,
            "data": self.sid.data,
            "links": self.sid_links,
            "stats": None,
        }
        self.assertSidUnchanged()

    @override_permission(Collection, "can_configure", AllowAll)
    def test_patch_rename(self) -> None:
        response = self.patch(data={"name": "renamed"})
        new_data = self.assertAPIResponseOk(response)
        self.sid.refresh_from_db()
        assert new_data == {
            "id": self.sid.pk,
            "category": self.sid.category,
            "name": "renamed",
            "full_history_retention_period": None,
            "metadata_only_retention_period": None,
            "data": self.sid.data,
            "links": self.get_links(self.sid),
            "stats": None,
        }
        self.assertEqual(self.sid.name, "renamed")
        self.assertEqual(self.sid.category, CollectionCategory.SUITE)
        self.assertIsNone(self.sid.full_history_retention_period)
        self.assertIsNone(self.sid.metadata_only_retention_period)
        self.assertEqual(self.sid.data, {"test": 1})

    @override_permission(Collection, "can_configure", AllowAll)
    def test_patch_change_retention_periods(self) -> None:
        response = self.patch(
            data={
                "full_history_retention_period": 123,
                "metadata_only_retention_period": 456,
            }
        )
        new_data = self.assertAPIResponseOk(response)
        assert new_data == {
            "id": self.sid.pk,
            "category": self.sid.category,
            "name": self.sid.name,
            "full_history_retention_period": 123,
            "metadata_only_retention_period": 456,
            "data": self.sid.data,
            "links": self.sid_links,
            "stats": None,
        }
        self.sid.refresh_from_db()
        self.assertEqual(self.sid.name, "sid")
        self.assertEqual(self.sid.category, CollectionCategory.SUITE)
        self.assertEqual(
            self.sid.full_history_retention_period,
            dt.timedelta(days=123),
        )
        self.assertEqual(
            self.sid.metadata_only_retention_period,
            dt.timedelta(days=456),
        )
        self.assertEqual(self.sid.data, {"test": 1})

    @override_permission(Collection, "can_configure", AllowAll)
    def test_patch_change_data(self) -> None:
        response = self.patch(data={"data": {"changed": True}})
        new_data = self.assertAPIResponseOk(response)
        assert new_data == {
            "id": self.sid.pk,
            "category": self.sid.category,
            "name": self.sid.name,
            "full_history_retention_period": None,
            "metadata_only_retention_period": None,
            "data": {"changed": True},
            "links": self.sid_links,
            "stats": None,
        }
        self.sid.refresh_from_db()
        self.assertEqual(self.sid.name, "sid")
        self.assertEqual(self.sid.category, CollectionCategory.SUITE)
        self.assertIsNone(self.sid.full_history_retention_period)
        self.assertIsNone(self.sid.metadata_only_retention_period)
        self.assertEqual(self.sid.data, {"changed": True})

    @override_permission(Collection, "can_configure", AllowAll)
    def test_patch_change_all(self) -> None:
        response = self.patch(
            data={
                "name": "renamed",
                "full_history_retention_period": 123,
                "metadata_only_retention_period": 456,
                "data": {"changed": True},
            }
        )
        new_data = self.assertAPIResponseOk(response)
        self.sid.refresh_from_db()
        assert new_data == {
            "id": self.sid.pk,
            "name": "renamed",
            "category": self.sid.category,
            "full_history_retention_period": 123,
            "metadata_only_retention_period": 456,
            "data": {"changed": True},
            "links": self.get_links(self.sid),
            "stats": None,
        }
        self.assertEqual(self.sid.name, "renamed")
        self.assertEqual(
            self.sid.full_history_retention_period,
            dt.timedelta(days=123),
        )
        self.assertEqual(
            self.sid.metadata_only_retention_period,
            dt.timedelta(days=456),
        )
        self.assertEqual(self.sid.data, {"changed": True})

    @override_permission(Collection, "can_configure", AllowAll)
    def test_patch_cannot_change_category(self) -> None:
        response = self.patch(
            data={
                "name": "renamed",
                "category": "debusine:test",
                "full_history_retention_period": 123,
                "metadata_only_retention_period": 456,
                "data": {"changed": True},
            }
        )
        self.assertResponseProblem(
            response,
            "The category of a collection cannot be changed",
            status_code=status.HTTP_400_BAD_REQUEST,
        )
        self.assertSidUnchanged()

    @override_permission(Collection, "can_configure", AllowAll)
    def test_patch_rename_conflict(self) -> None:
        self.playground.create_collection("renamed", CollectionCategory.SUITE)

        response = self.patch(data={"name": "renamed"})
        self.assertResponseProblem(
            response,
            "New name conflicts with an existing collection",
            detail_pattern=(
                "A collection with category 'debian:suite'"
                " called 'renamed' already exists"
            ),
            status_code=status.HTTP_400_BAD_REQUEST,
        )
        self.assertSidUnchanged()

    @override_permission(Collection, "can_configure", AllowAll)
    def test_patch_rename_no_conflict_in_other_workspaces(self) -> None:
        other = self.playground.create_workspace(name="other")
        self.playground.create_collection(
            "renamed", CollectionCategory.SUITE, workspace=other
        )

        response = self.patch(data={"name": "renamed"})
        new_data = self.assertAPIResponseOk(response)
        self.sid.refresh_from_db()
        assert new_data == {
            "id": self.sid.pk,
            "category": self.sid.category,
            "name": "renamed",
            "full_history_retention_period": (
                self.sid.full_history_retention_period
            ),
            "metadata_only_retention_period": (
                self.sid.metadata_only_retention_period
            ),
            "data": self.sid.data,
            "links": self.get_links(self.sid),
            "stats": None,
        }

    @override_permission(Collection, "can_configure", AllowAll)
    def test_patch_rename_singleton_ok(self) -> None:
        collection = self.scenario.workspace.get_singleton_collection(
            pc=self.scenario.context_ui(), category=CollectionCategory.ARCHIVE
        )
        response = self.patch(
            collection=collection,
            data={"name": collection.name},
        )
        new_data = self.assertAPIResponseOk(response)
        assert new_data == {
            "id": collection.pk,
            "category": collection.category,
            "name": collection.name,
            "full_history_retention_period": (
                collection.full_history_retention_period
            ),
            "metadata_only_retention_period": (
                collection.metadata_only_retention_period
            ),
            "data": collection.data,
            "links": self.get_links(collection),
            "stats": None,
        }

    @override_permission(Collection, "can_configure", AllowAll)
    def test_patch_rename_singleton_invalid_name(self) -> None:
        collection = self.scenario.workspace.get_singleton_collection(
            pc=self.scenario.context_ui(), category=CollectionCategory.ARCHIVE
        )
        response = self.patch(
            collection=collection,
            data={"name": "renamed"},
        )
        self.assertResponseProblem(
            response,
            "Invalid name for singleton collection",
            detail_pattern=(
                "'debian:archive' singleton collections must be named '_'"
            ),
            status_code=status.HTTP_400_BAD_REQUEST,
        )
        collection.refresh_from_db()
        self.assertEqual(collection.name, "_")

    @override_permission(Collection, "can_delete", AllowAll)
    def test_delete(self) -> None:
        response = self.delete(collection=self.sid)
        new_data = self.assertAPIResponseOk(
            response, status_code=status.HTTP_204_NO_CONTENT
        )
        self.assertIsNone(new_data)
        self.assertFalse(Collection.objects.filter(pk=self.sid.pk).exists())
        self.assertTrue(Collection.objects.filter(pk=self.trixie.pk).exists())

    def test_debusine_collection_show(self) -> None:
        debusine = self.get_debusine()
        result = debusine.collection_get(
            self.scenario.workspace.name, CollectionCategory.SUITE, "sid"
        )
        self.assertIsNone(result.stats)
        result = debusine.collection_get(
            self.scenario.workspace.name,
            CollectionCategory.SUITE,
            "sid",
            with_stats=True,
        )
        self.assertEqual(result.stats, [])

    def test_debusine_collection_iter(self) -> None:
        debusine = self.get_debusine()
        result = debusine.collection_iter(self.scenario.workspace.name)
        self.assertCountEqual(
            [(c.category, c.name) for c in result],
            [
                (CollectionCategory.PACKAGE_BUILD_LOGS, "_"),
                (CollectionCategory.TASK_HISTORY, "_"),
                (CollectionCategory.TASK_CONFIGURATION, "default"),
                (CollectionCategory.ARCHIVE, "_"),
                (CollectionCategory.SUITE, "sid"),
                (CollectionCategory.SUITE, "trixie"),
            ],
        )

    def test_debusine_collection_get(self) -> None:
        debusine = self.get_debusine()
        result = debusine.collection_get(
            self.scenario.workspace.name, self.sid.category, self.sid.name
        )
        self.assertEqual(result.id, self.sid.id)
        self.assertEqual(result.name, self.sid.name)
        self.assertEqual(result.category, self.sid.category)
        self.assertEqual(result.data, self.sid.data)

    def test_debusine_collection_get_not_found(self) -> None:
        debusine = self.get_debusine()
        with self.assertRaisesRegex(DebusineError, r"Object not found"):
            debusine.collection_get(
                self.scenario.workspace.name,
                self.sid.category,
                "does-not-exist",
            )

    @override_permission(Workspace, "can_create_collections", AllowAll)
    def test_debusine_create_collection(self) -> None:
        debusine = self.get_debusine()
        result = debusine.collection_create(
            self.scenario.workspace.name,
            CollectionDataNew(
                name="new",
                category=CollectionCategory.SUITE,
                full_history_retention_period=12,
                data={"created": True},
            ),
        )
        new = Collection.objects.get(
            workspace=self.scenario.workspace,
            category=CollectionCategory.SUITE,
            name="new",
        )
        self.assertEqual(result.id, new.pk)
        self.assertEqual(result.name, "new")
        self.assertEqual(result.category, CollectionCategory.SUITE)
        self.assertEqual(result.full_history_retention_period, 12)
        self.assertIsNone(result.metadata_only_retention_period)
        self.assertEqual(result.data, {"created": True})

        self.assertEqual(
            new.full_history_retention_period, dt.timedelta(days=12)
        )
        self.assertIsNone(new.metadata_only_retention_period)
        self.assertEqual(new.data, {"created": True})

    @override_permission(Collection, "can_configure", AllowAll)
    def test_debusine_collection_update(self) -> None:
        debusine = self.get_debusine()
        result = debusine.collection_update(
            self.scenario.workspace.name,
            CollectionData(
                id=self.sid.id,
                name=self.sid.name,
                category=self.sid.category,
                full_history_retention_period=3,
                data=self.sid.data,
            ),
        )
        self.assertEqual(result.id, self.sid.id)
        self.assertEqual(result.name, self.sid.name)
        self.assertEqual(result.category, self.sid.category)
        self.assertEqual(result.full_history_retention_period, 3)
        self.assertEqual(result.data, self.sid.data)
        self.sid.refresh_from_db()
        self.assertEqual(
            self.sid.full_history_retention_period, dt.timedelta(days=3)
        )

    @override_permission(Collection, "can_delete", AllowAll)
    def test_debusine_collection_delete(self) -> None:
        debusine = self.get_debusine()
        debusine.collection_delete(
            self.scenario.workspace.name,
            CollectionData(
                id=self.sid.id,
                name=self.sid.name,
                category=self.sid.category,
                full_history_retention_period=3,
                data=self.sid.data,
            ),
        )
        self.assertFalse(Collection.objects.filter(pk=self.sid.pk).exists())

    @override_permission(Collection, "can_delete", DenyAll)
    def test_debusine_collection_delete_not_allowed(self) -> None:
        debusine = self.get_debusine()
        with self.assertRaisesRegex(
            DebusineError,
            "playground cannot delete collection sid@debian:suite",
        ):
            debusine.collection_delete(
                self.scenario.workspace.name,
                CollectionData(
                    id=self.sid.id,
                    name=self.sid.name,
                    category=self.sid.category,
                    full_history_retention_period=3,
                    data=self.sid.data,
                ),
            )
        self.assertTrue(Collection.objects.filter(pk=self.sid.pk).exists())


class CollectionTokenGrantViewSetTests(TestCase):
    """Tests for :py:class:`CollectionTokenGrantViewSet`."""

    scenario = scenarios.DefaultContextAPI()
    sid: ClassVar[Collection]
    tokens: ClassVar[list[Token]]
    grants: ClassVar[list[CollectionTokenGrant]]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.sid = cls.playground.create_collection(
            "sid", CollectionCategory.SUITE
        )
        cls.tokens = [
            cls.playground.create_external_user_token(
                username=username, comment=comment
            )
            for username, comment in (("test1", ""), ("test2", "Test comment"))
        ]
        cls.grants = [
            cls.sid.grant_role(
                token, Collection.Roles.VIEWER, created_by=cls.scenario.user
            )
            for token in cls.tokens
        ]

    def _collection_url_kwargs(
        self, collection: Collection | str | None
    ) -> dict[str, str]:
        match collection:
            case None:
                workspace_name = self.sid.workspace.name
                collection_name = self.sid.name
                collection_category = self.sid.category
            case str():
                workspace_name, collection_lookup = collection.split("/", 1)
                assert "@" in collection_lookup
                collection_name, collection_category = collection_lookup.rsplit(
                    "@", 1
                )
            case _:
                workspace_name = collection.workspace.name
                collection_name = collection.name
                collection_category = collection.category
        return {
            "workspace": workspace_name,
            "ccat": collection_category,
            "cname": collection_name,
        }

    def _reverse_list(
        self, collection: Collection | str | None, **kwargs: str
    ) -> str:
        """Build the URL for a test request."""
        kwargs.update(self._collection_url_kwargs(collection))
        return reverse("api:collection-token-grant-list", kwargs=kwargs)

    def _reverse_detail(
        self,
        grant: (
            CollectionTokenGrant | tuple[str, Collection.Roles] | None
        ) = None,
        collection: Collection | str | None = None,
        **kwargs: str,
    ) -> str:
        """Build the URL for a test request."""
        match grant:
            case None:
                kwargs.update(
                    self._collection_url_kwargs(collection or self.sid)
                )
                kwargs["token"] = self.grants[0].token.hash
                kwargs["role"] = self.grants[0].role
            case (str() as token_hash, Collection.Roles() as role):
                kwargs.update(
                    self._collection_url_kwargs(collection or self.sid)
                )
                kwargs["token"] = token_hash
                kwargs["role"] = role
            case CollectionTokenGrant():
                kwargs.update(
                    self._collection_url_kwargs(collection or grant.collection)
                )
                kwargs["token"] = grant.token.hash
                kwargs["role"] = grant.role
            case _ as unreachable:
                raise AssertionError(f"Unexpected grant: {unreachable}")
        return reverse("api:collection-token-grant-detail", kwargs=kwargs)

    def _headers(self, user_token: bool) -> dict[str, str]:
        """Build headers for a test request."""
        match user_token:
            case True:
                return {"token": self.scenario.user_token.key}
            case False:
                return {}
            case _ as unreachable:
                assert_never(unreachable)

    def list(
        self,
        collection: Collection | str | None = None,
        user_token: bool = True,
    ) -> TestResponseType:
        """GET request for all token grants."""
        return self.client.get(
            self._reverse_list(collection=collection),
            content_type="application/json",
            headers=self._headers(user_token),
        )

    def get(
        self,
        grant: (
            CollectionTokenGrant | tuple[str, Collection.Roles] | None
        ) = None,
        collection: Collection | str | None = None,
        user_token: bool = True,
    ) -> TestResponseType:
        """GET request for a particular token grant."""
        return self.client.get(
            self._reverse_detail(grant=grant, collection=collection),
            content_type="application/json",
            headers=self._headers(user_token),
        )

    def post(
        self,
        collection: Collection | str | None = None,
        user_token: bool = True,
        data: dict[str, Any] | None = None,
    ) -> TestResponseType:
        """POST request."""
        return self.client.post(
            self._reverse_list(collection=collection),
            data=data or {},
            content_type="application/json",
            headers=self._headers(user_token),
        )

    def delete(
        self,
        grant: (
            CollectionTokenGrant | tuple[str, Collection.Roles] | None
        ) = None,
        collection: Collection | str | None = None,
        user_token: bool = True,
    ) -> TestResponseType:
        """DELETE request for a named collection."""
        return self.client.delete(
            self._reverse_detail(grant=grant, collection=collection),
            headers=self._headers(user_token),
        )

    def test_unauthenticated(self) -> None:
        """Authentication is required."""
        for method in ("list", "get", "post", "delete"):
            with self.subTest(method=method):
                response = getattr(self, method)(user_token=False)
                self.assertResponseProblem(
                    response,
                    "Error",
                    detail_pattern=(
                        "Authentication credentials were not provided."
                    ),
                    status_code=status.HTTP_403_FORBIDDEN,
                )

    def test_workspace_not_found(self) -> None:
        """Workspace must exist."""
        for method in ("list", "get", "post", "delete"):
            with self.subTest(method=method):
                response = getattr(self, method)(
                    collection=f"does-not-exist/sid@{CollectionCategory.SUITE}"
                )
                self.assertResponseProblem(
                    response,
                    "Workspace not found",
                    detail_pattern=(
                        "Workspace does-not-exist not found in scope debusine"
                    ),
                    status_code=status.HTTP_404_NOT_FOUND,
                )

    def test_workspace_not_visible(self) -> None:
        """User must be able to view the workspace."""
        workspace = self.playground.create_workspace(name="private")
        collection = self.playground.create_collection(
            name="test", category=CollectionCategory.TEST, workspace=workspace
        )
        for method in ("list", "get", "post", "delete"):
            with self.subTest(method=method):
                response = getattr(self, method)(collection=collection)
                self.assertResponseProblem(
                    response,
                    "Workspace not found",
                    detail_pattern=(
                        "Workspace private not found in scope debusine"
                    ),
                    status_code=status.HTTP_404_NOT_FOUND,
                )

    def test_collection_not_found(self) -> None:
        """Collection must exist."""
        collection_name = f"nonexistent@{CollectionCategory.SUITE}"
        for method in ("list", "get", "post", "delete"):
            with self.subTest(method=method):
                response = getattr(self, method)(
                    collection=(
                        f"{self.scenario.workspace.name}/{collection_name}"
                    )
                )
                self.assertResponseProblem(
                    response,
                    "Collection not found",
                    detail_pattern=(
                        f"Collection {collection_name!r} does not exist in"
                        f" workspace {self.scenario.workspace}"
                    ),
                    status_code=status.HTTP_404_NOT_FOUND,
                )

    def test_collection_not_visible(self) -> None:
        """User must be able to view the collection."""
        for method in ("list", "get", "post", "delete"):
            with (
                self.subTest(method=method),
                override_permission(Collection, "can_display", DenyAll),
            ):
                response = getattr(self, method)()
                self.assertResponseProblem(
                    response,
                    f"{self.scenario.user} cannot display collection"
                    f" {self.sid.name}@{self.sid.category}",
                    status_code=status.HTTP_403_FORBIDDEN,
                )

    def test_grant_nonexistent_hash(self) -> None:
        """The grant's token hash must exist."""
        for method in ("get", "delete"):
            with self.subTest(method=method):
                response = getattr(self, method)(
                    grant=("nonexistent", Collection.Roles.VIEWER)
                )
                self.assertResponseProblem(
                    response,
                    "Object not found",
                    status_code=status.HTTP_404_NOT_FOUND,
                )

    def test_grant_nonexistent_role(self) -> None:
        """The grant's role must exist."""
        for method in ("get", "delete"):
            with self.subTest(method=method):
                response = getattr(self, method)(
                    grant=(self.grants[0].token.hash, Collection.Roles.OWNER)
                )
                self.assertResponseProblem(
                    response,
                    "Object not found",
                    status_code=status.HTTP_404_NOT_FOUND,
                )

    def test_create_unauthorized(self) -> None:
        """The user must be able to create the grant."""
        with override_permission(Collection, "can_configure", DenyAll):
            response = self.post(
                data={"username": "test", "role": Collection.Roles.VIEWER}
            )
            self.assertResponseProblem(
                response,
                f"{self.scenario.user} cannot configure collection"
                f" {self.sid.name}@{self.sid.category}",
                status_code=status.HTTP_403_FORBIDDEN,
            )

    def test_delete_unauthorized(self) -> None:
        """The user must be able to delete the grant."""
        with override_permission(Collection, "can_configure", DenyAll):
            response = self.delete()
            self.assertResponseProblem(
                response,
                f"{self.scenario.user} cannot configure collection"
                f" {self.sid.name}@{self.sid.category}",
                status_code=status.HTTP_403_FORBIDDEN,
            )

    def test_list(self) -> None:
        response = self.list()
        data = self.assertAPIResponseOk(response)
        self.assertEqual(
            data["results"],
            [
                {
                    "token_type": Token.Types.EXTERNAL_USER,
                    "token_hash": self.tokens[0].hash,
                    "username": "test1",
                    "role": Collection.Roles.VIEWER,
                    "comment": "",
                },
                {
                    "token_type": Token.Types.EXTERNAL_USER,
                    "token_hash": self.tokens[1].hash,
                    "username": "test2",
                    "role": Collection.Roles.VIEWER,
                    "comment": "Test comment",
                },
            ],
        )

    def test_detail(self) -> None:
        response = self.get()
        data = self.assertAPIResponseOk(response)
        self.assertEqual(
            data,
            {
                "token_type": Token.Types.EXTERNAL_USER,
                "token_hash": self.tokens[0].hash,
                "username": "test1",
                "role": Collection.Roles.VIEWER,
                "comment": "",
            },
        )

    @override_permission(Collection, "can_configure", AllowAll)
    def test_post_token_hash(self) -> None:
        token = self.playground.create_external_user_token(username="test-new")
        response = self.post(
            data={"token_hash": token.hash, "role": Collection.Roles.VIEWER}
        )
        new_data = self.assertAPIResponseOk(
            response, status_code=status.HTTP_201_CREATED
        )
        self.assertEqual(
            new_data,
            {
                "token_type": Token.Types.EXTERNAL_USER,
                "token_hash": token.hash,
                "username": "test-new",
                "role": Collection.Roles.VIEWER,
                "comment": "",
            },
        )
        self.assertTrue(
            token.collection_grants.filter(
                collection=self.sid, role=Collection.Roles.VIEWER
            ).exists()
        )

    @override_permission(Collection, "can_configure", AllowAll)
    def test_post_username(self) -> None:
        response = self.post(
            data={
                "username": "test-new",
                "role": Collection.Roles.VIEWER,
                "comment": "Very important user",
            }
        )
        new_data = self.assertAPIResponseOk(
            response, status_code=status.HTTP_201_CREATED
        )
        token_key = new_data.pop("token_key")
        token = Token.objects.get_token_or_none(token_key)
        assert token is not None
        self.assertEqual(
            new_data,
            {
                "token_type": Token.Types.EXTERNAL_USER,
                "token_hash": token.hash,
                "username": "test-new",
                "role": Collection.Roles.VIEWER,
                "comment": "Very important user",
            },
        )
        self.assertTrue(token.enabled)
        self.assertTrue(
            token.collection_grants.filter(
                collection=self.sid, role=Collection.Roles.VIEWER
            ).exists()
        )

    @override_permission(Collection, "can_configure", AllowAll)
    def test_post_invalid_username(self) -> None:
        response = self.post(
            data={"username": "foo bar&baz", "role": Collection.Roles.VIEWER}
        )
        self.assertResponseProblem(
            response,
            "Cannot grant collection token",
            status_code=status.HTTP_400_BAD_REQUEST,
            validation_errors_pattern="Enter a valid username",
        )

    @override_permission(Collection, "can_configure", AllowAll)
    def test_delete(self) -> None:
        response = self.delete(grant=self.grants[0])
        new_data = self.assertAPIResponseOk(
            response, status_code=status.HTTP_204_NO_CONTENT
        )
        self.assertIsNone(new_data)
        self.assertFalse(
            self.sid.token_grants.filter(pk=self.grants[0].pk).exists()
        )
        self.assertTrue(
            self.sid.token_grants.filter(pk=self.grants[1].pk).exists()
        )
